# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
HTML charts.
'''

from functools import reduce
from math import floor, log10
from typing import Any, Callable, Dict, Iterable, Optional, Sequence, Tuple, Union, overload

from ..range import Num, NumRange
from . import Div, HtmlNode, Span, MuAttrs, InlineStyle


Dim = Union[int, float, str]
Vec = Tuple[Num, Num]
VecOrNum = Union[Vec, Num]
F2 = Tuple[float, float]
F2OrF = Union[F2, float]
BoundsF2 = Tuple[F2, F2]
PathCommand = Tuple

PointTransform = Callable[[Tuple], F2]
TickFmt = Callable[[float], Any]
Plotter = Callable[[Div, PointTransform, Any], None]



class ChartSeries:

  name:str
  legend:str
  bounds:Optional[BoundsF2] = None # Overridden by subclasses.

  def render(self, chart:'HtmlChart', container:Div) -> None: raise NotImplementedError



class ChartAxis:

  def __init__(self, *,
   min:Optional[Num]=None, max:Optional[Num]=None,
   visible_origin=False,
   symmetric=False,
   show_grid=True, grid:Iterable[Num]=(), grid_step:Num=0.0, grid_min:Num=16,
   show_ticks=True, ticks:Iterable[Num]=(), tick_step:Num=0.0, tick_space:Num=1,
   tick_w:Num=16, tick_fmt:Optional[TickFmt]=None) -> None:

    self._min = None if min is None else float(min)
    self._max = None if max is None else float(max)
    self.visible_origin = visible_origin
    self.symmetric = symmetric
    self.show_grid = show_grid
    self.grid = list(grid)
    self.grid_step = float(grid_step)
    self.grid_min = float(grid_min)
    self.show_ticks = show_ticks
    self.ticks = list(ticks)
    self.tick_step = float(tick_step)
    self.tick_space = float(tick_space)
    self.tick_w = float(tick_w)
    self.tick_fmt = tick_fmt

    # These attributes are filled in later.
    self.idx = -1
    self.data_len = 0.0
    self.min = 0.0
    self.max = 0.0
    self.transform = _axis_transform_dummy
    self.scale = 0.0

  def calc_min_max(self, data_bounds:Optional[BoundsF2]) -> None:
    # Determine data data_bounds.
    _min:Optional[float] = self._min
    _max:Optional[float] = self._max
    if _min is None: _min = 0.0   if data_bounds is None else data_bounds[0][self.idx]
    if _max is None: _max = _min if data_bounds is None else data_bounds[1][self.idx]
    if _max <= _min: _max = _min + 1.0
    if self.visible_origin:
      if _min > 0.0:   _min = 0.0
      elif _max < 0.0: _max = 0.0
    if self.symmetric:
      _max = max(_max, -_min)
      _min = -_max
    self.min = _min
    self.max = _max

  def calc_layout(self, chart_size:F2, title_h:float, tick_len:float, tick_h:float) -> None:
    self.data_len = self.max - self.min
    self.scale = 1 / self.data_len

    # Calculate tick step.
    tick_min_screen_step = (self.tick_w*1.5) if (self.idx == 0) else (tick_h*2.0)
    tick_mult = 1.0
    frac_w = 0
    fmt_w = 0
    if self.tick_step <= 0 and tick_min_screen_step > 0:
      step1, tick_mult = self.choose_step(tick_min_screen_step)
      self.tick_step = step1 * tick_mult
    if self.tick_step > 0 and not self.tick_fmt:
      exp = floor(log10(self.tick_step))
      frac_w = max(0, -exp)
      fmt_w = max(
        len(f'{self.min:,.{frac_w}f}'),
        len(f'{self.max:,.{frac_w}f}'))

      pad_str = f'{10**fmt_w:,.{frac_w}f}'[-fmt_w:] # Longer than necessary. Take pad chars from right to left.
      def tick_fmt(val:float) -> Any:
        s = f'{val:,.{frac_w}f}'
        return Span(cl='zpad', ch=pad_str[:-len(s)]), s

      self.tick_fmt = tick_fmt

    # Calculate grid step.
    if self.grid_step <= 0:
      step1, mult = self.choose_step(min_screen_step=self.grid_min)
      if mult == 2 and tick_mult == 5: # Ticks will misalign to grid; bump grid to 2.5.
        mult = 2.5
      self.grid_step = step1 * mult

    if self.ticks:
      if not self.grid:
        self.grid = self.ticks
    else: # Calculate ticks.
      ti, tr = divmod(self.min, self.tick_step)
      if tr > 0.1: ti += 1 # If the remainder is visually significant, skip the first tick.
      t_start = ti*self.tick_step
      self.ticks = [t for t in NumRange(t_start, self.max, step=self.tick_step, closed=True)]
      if not self.grid: # Calculate grid.
        g_start = (self.min//self.grid_step + 1) * self.grid_step # Skip line index 0 because it is always <= low border.
        self.grid = [g for g in NumRange(g_start, self.max, self.grid_step)]



  def choose_step(self, min_screen_step:float) -> Tuple[float, float]:
    assert self.data_len > 0
    assert min_screen_step > 0
    axis_length = 100 # TEMP
    cram_num = max(1.0, axis_length // min_screen_step) # Maximum number of ticks that could be placed.
    assert cram_num > 0, (cram_num, axis_length, min_screen_step)
    cram_step = self.data_len / cram_num # Minimum fractional data step.
    exp = floor(log10(cram_step))
    step1 = float(10**exp) # Low estimate of step.
    for mult in (1.0, 2.0, 5.0):
      step = step1 * mult
      if step >= cram_step: return step1, mult
    return step1, 10.0



class HtmlChart(HtmlNode):
  tag = 'div'

  def __init__(self,
   title:str=None,
   x:ChartAxis=None,
   y:ChartAxis=None,
   series:Sequence[ChartSeries]=(),
   symmetric_xy=False,
   dbg=False,
   attrs:MuAttrs=None) -> None:

    attrs = attrs or {}
    super().__init__(attrs=attrs)

    self.prepend_class('chart')
    self.title = title
    self.x = x = ChartAxis() if x is None else x
    self.y = y = ChartAxis() if y is None else y
    self.series = series
    self.symmetric_xy = symmetric_xy

    x.idx = 0
    y.idx = 1

    data_bounds = reduce(expand_opt_bounds, (s.bounds for s in series), None)

    x.calc_min_max(data_bounds=data_bounds)
    y.calc_min_max(data_bounds=data_bounds)

    if symmetric_xy:
      x.min = min(x.min, y.min)
      y.min = x.min
      x.max = max(x.max, y.max)
      y.max = x.max

    # Calculate per-axis length as necessary.
    size = (960,480) # TEMP
    title_h = 14 # TEMP
    tick_len = 48 # TEMP
    tick_h = 16 # TEMP
    x.calc_layout(chart_size=size, title_h=title_h, tick_len=tick_len, tick_h=tick_h)
    y.calc_layout(chart_size=size, title_h=title_h, tick_len=tick_len, tick_h=tick_h)

    self.data_size = data_size = (x.data_len, y.data_len)
    self.grid_size = grid_size = (100, 100) # TEMP
    self.scale = (x.scale, y.scale)

    assert x is not None
    assert y is not None
    ax = x # Hack around mypy.
    ay = y

    def transform(point:Sequence) -> F2:
      'Translate a point to from data space to div space.'
      px = float(point[0])
      py = float(point[1])
      return (round(ax.scale*(px-ax.min), 1), round(ay.scale*(ay.data_len - (py-ay.min)), 1))

    def x_transform(val:Num) -> float: return round(ax.scale*(float(val) - ax.min), 1)
    def y_transform(val:Num) -> float: return round(ay.scale*(ay.data_len - (float(val)-ay.min)), 1)


    self.transform = transform
    x.transform = x_transform
    y.transform = y_transform

    # Contents.

    # Title.
    if self.title is not None:
      self.append(Span(cl='title', ch=self.title))

    area = self.append(Div(cl='chart-area')) # Leave area open.

    # Clip path is is defined to match grid.
    #clip_path_id = area.gen_id()
    #self.chart_clip_path = f'url(#{clip_path_id})'
    #with area.clipPath(id=clip_path_id) as clipPath:
    #  clipPath.rect(size=grid_size, r=corner_radius)

    # Grid.
    if x.show_grid:
      g_start_x = (x.min//x.grid_step + 1) * x.grid_step # Skip line index 0 because it is always <= low border.
      for gx in x.grid: # X axis.
        tgx = x.transform(gx)
        #grid.line((tgx, 0), (tgx, y.length)) # Vertical lines.
    if y.show_grid:
      g_start_y = (y.min//y.grid_step + 1) * y.grid_step # Skip line index 0 because it is always <= low border.
      for gy in y.grid:
        tgy = y.transform(gy)
        #grid.line((0, tgy), (x.length, tgy)) # Horizontal lines.

    # Axes.
    if y.min <= 0 and y.max >= 0: # Draw X axis.
      y0 = y.transform(0)
      #area.line((0, y0), (x.length, y0), cl='axis', id='x-axis')
    if x.min <= 0 and x.max >= 0: # Draw Y axis.
      x0 = x.transform(0)
      #area.line((x0, 0), (x0, y.length), cl='axis', id='y-axis')

    def handle_rendered_tick(val:Any) -> Tuple:
      if isinstance(val, str): return (val,)
      try: return tuple(val)
      except TypeError: return (val,)

    # Ticks.
    if x.show_ticks:
      tick_x = area.append(Div(cl='tick-x'))
      txi, txr = divmod(x.min, x.tick_step)
      if txr > 0.1: txi += 1 # If the remainder is visually significant, skip the first tick.
      t_start_x = txi*x.tick_step
      for _x in x.ticks:
        tx = x.transform(_x)
        ty = 100.0 # TEMP
        tb = ty + tick_len
        tty = tb + x.tick_space
        #tick_x.line((tx, ty), (tx, tb), cl='tick')
        assert x.tick_fmt is not None
        #tick_x.append(Span(ch=str(x.tick_fmt(_x)), pos=(tx, tty), cl='tick'))
    if y.show_ticks:
      tick_y = area.append(Div(cl='tick-y'))
      tyi, tyr = divmod(y.min, y.tick_step)
      if tyr > 0.1: tyi += 1 # If the remainder is visually significant, skip the first tick.
      t_start_y = tyi*y.tick_step
      for _y in y.ticks:
        tx = 100.0 # TEMP
        tr = tx + tick_len
        ttx = tr + y.tick_space
        ty = y.transform(_y)
        #tick_y.line((tx, ty), (tr, ty), cl='tick')
        assert y.tick_fmt is not None
        # tick_y.append(Span(ch=str(y.tick_fmt(_y)), pos=(ttx, ty), cl='tick'))

    # Legend.
    legend_w = 200
    legend_h = 50
    if legend_h > 0:
      leg_y = size[1] - title_h - legend_h - 1
      legend_g = self.append(Div(cl='chart-legend'))
      for i, s in enumerate(series):
        g = legend_g.append(Div(cl='chart-legend-'+s.name, ch=s.legend))


    # Series.
    series_container = area.append(Div(cl='series'))
    for s in series:
      series_div = series_container.append(Div(cl=s.name))
      s.render(self, container=series_div)



class BarSeries(ChartSeries):

  def __init__(self, name:str, points:Sequence[Tuple], numeric:bool, legend:str='', width=1.0, plotter:Optional[Plotter]=None,
   title_fmt:Optional[Callable[[Tuple], str]]=None, **attrs:Any) -> None:

    self.name = name
    self.points = list(points)
    self.numeric = numeric
    self.legend = legend or name
    self.width = width
    self.plotter = plotter
    self.title_fmt = title_fmt
    self.attrs = attrs
    self.bounds:Optional[Tuple[F2, F2]] = None

    def float_from(val:Any, label:str) -> float:
      try: return float(val)
      except TypeError as e: raise TypeError(f'BarSeries received non-numeric point {label}: {val!r}') from e

    if self.points:
      for p in self.points: # Get first value.
        x = p[0]
        y = p[1]
        if numeric:
          x_min = x_max = float_from(x, 'key')
        y_min = y_max = float_from(y, 'value')
        break
      for p in self.points:
        x = p[0]
        y = p[1]
        if numeric:
          x = float_from(x, 'key')
          x_min = min(x_min, x)
          x_max = max(x_max, x)
        y = float_from(y, 'value')
        y_min = min(y_min, y)
        y_max = max(y_max, y)
      if numeric:
        half_w = self.width * 0.5
        self.bounds = ((x_min - half_w, 0.0), (x_max + half_w, y_max))
      else:
        self.bounds = ((0.0, 0.0), (float(len(self.points)), y_max))


  def render(self, chart:'HtmlChart', container:Div) -> None:
    assert self.plotter is None # TODO
    w = chart.x.scale * self.width
    # Do not place the bars in a group because we want to be able to z-index bars across multiple series.
    for i, p in enumerate(self.points):
      if not self.numeric:
        p = (i+0.5, p[1], p[0])
      (x_mid, y) = chart.transform(p)
      assert p[1] >= 0, f'negative bar values are not yet supported: {p!r}'
      kwargs:Dict[str,Any] = {}
      if self.title_fmt:
        kwargs['title'] = self.title_fmt(p)
      left = x_mid - w*0.5
      right = x_mid + w*0.5
      bottom = 0
      top = y
      style = InlineStyle(
        left=f'{left:.3%}',
        right=f'{right:.3%}',
        bottom=f'{bottom:.3%}',
        top=f'{top:.3%}',
        z_index=i)
      container.append(Div(style=style, **kwargs, **self.attrs)) # TODO: Custom format of title?



def _axis_transform_dummy(val:Num) -> float: raise Exception('missing transform')
def scale(x:Num=1, y:Num=1) -> str: return f'scale({x},{y})'
def rotate(degrees:Num, x:Num=0, y:Num=0) -> str: return f'rotate({fmt_num(degrees)},{fmt_num(x)},{fmt_num(y)})'
def translate(x:Num=0, y:Num=0) -> str: return f'translate({fmt_num(x)},{fmt_num(y)})'


@overload
def fmt_num(n:Num) -> str: ...
@overload
def fmt_num(n:None) -> None: ...

def fmt_num(n:Optional[Num]) -> Optional[str]:
  'Remove trailing ".0" from floats that can be represented as integers.'
  if n is None: return None
  if isinstance(n, float):
    i = int(n)
    if i == n: return str(i)
  return str(n)


def f2_for_vec(v:Vec) -> F2:
  x, y = v
  return (float(x), float(y))


def unpack_VecOrNum(vn:VecOrNum) -> Tuple[float, float]:
  if isinstance(vn, tuple):
    x, y = vn
    return float(x), float(y)
  else:
    s = float(vn)
    return (s, s)


def expand_opt_bounds(l:Optional[BoundsF2], r:Optional[BoundsF2]) -> Optional[BoundsF2]:
  if l is None: return r
  if r is None: return l
  (llx, lly), (lhx, lhy) = l
  (rlx, rly), (rhx, rhy) = r
  return ((min(llx, rlx), min(lly, rly)), (max(lhx, rhx), max(lhy, rhy)))


chart_style = '''
.chart {
  background-color: red;
  width: 100%;
  height: 16rem;
}
.chart > .chart-legend {
  background-color: blue;
}
.chart > div.chart-area {
  background-color: yellow;
  width: 100%;
  height: 100%;
}
.chart .series {
  position: relative;
  background-color: green;
}

.chart .series div {
  position: absolute;
}

text {
  stroke: none;
  fill: currentColor;
}
text.title {
  text-anchor: start;
  alignment-baseline: hanging;
}
g.tick-x text.tick {
  white-space: pre;
  text-anchor: start;
  alignment-baseline: hanging;
}
g.tick-y text.tick {
  white-space: pre;
  text-anchor: start;
  alignment-baseline: alphabetic;
}
g.chart-legend text {
  text-anchor: start;
  alignment-baseline: central;
}
'''



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'String utilities.'

import re
from decimal import Decimal
from string import Template
from typing import Any, Iterable, Iterator, Sequence, Tuple, TypeVar

_T = TypeVar('_T')


def render_template(template:str, **substitutions:Any) -> str:
  'Render a template using $ syntax.'
  t = Template(template)
  return t.substitute(substitutions)


def string_contains(string:str, query:str) -> bool:
  'Return True if string contains query.'
  return string.find(query) != -1


def clip_prefix(string:str, prefix:str, req=True) -> str:
  'Remove `prefix` if it is present, or raise ValueError, unless `req` is False.'
  if string.startswith(prefix):
    return string[len(prefix):]
  elif req:
    raise ValueError(string)
  return string


def clip_suffix(string:str, suffix:str, req=True) -> str:
  'Remove `suffix`if it is present, or raise ValueError, unless `req` is False.'
  if len(suffix) == 0: return string # need this case because string[:-0] == ''.
  if string.endswith(suffix):
    return string[:-len(suffix)]
  elif req:
    raise ValueError(string)
  return string


def replace_prefix(string:str, prefix:str, replacement:str, req=True) -> str:
  'Replace `prefix` if it is present, or raise ValueError, unless `req` is False.'
  if string.startswith(prefix):
    return replacement + string[len(prefix):]
  elif req:
    raise ValueError(string)
  return string


def replace_suffix(string:str, suffix:str, replacement:str, req=True) -> str:
  'Replace `suffix`if it is present, or raise ValueError, unless `req` is False.'
  if len(suffix) == 0: return string # need this case because string[:-0] == ''.
  if string.endswith(suffix):
    return string[:-len(suffix)] + replacement
  elif req:
    raise ValueError(string)
  return string


def clip_first_prefix(string:str, prefixes:Sequence[str], req=True) -> str:
  'Remove the first matching prefix in `prefixes` from `string`, or raise ValueError, unless `req is False`.'
  for p in prefixes:
    try:
      return clip_prefix(string, p, req=True)
    except ValueError:
      continue
  if req: raise ValueError(string)
  else: return string


def clip_common(strings:Sequence[str], prefix=True, suffix=True) -> Tuple[str,...]:
  if not strings: return ()
  if len(strings) == 1: return ('',)
  first = strings[0]
  min_len = min(len(s) for s in strings)
  i = 0
  if prefix:
    while i < min_len and all(s[i] == first[i] for s in strings):
      i += 1
  l = -1 # Last index.
  if suffix:
    l_term = i - min_len # The final negative index before we would clip everything from `i` to end.
    while l > l_term and all(s[l] == first[l] for s in strings):
      l -= 1
  return tuple(s[i:l] for s in strings)


def replace_first_prefix(string:str, prefixes:Sequence[Tuple[str,str]], req=True) -> str:
  'Replace the first matching prefix in `prefixes` from `string`, or raise ValueError, unless `req is False`.'
  for prefix, replacement in prefixes:
    try:
      return replace_prefix(string, prefix, replacement, req=True)
    except ValueError:
      continue
  if req:
    raise ValueError(string)
  return string


def find_and_clip_suffix(string:str, suffix:str, req=True) -> str:
  idx = string.find(suffix)
  if idx == -1:
    if req: raise ValueError(string)
    else: return string
  return string[:idx]


def indent_lines(lines:Iterable[str], depth=1) -> Iterator[str]:
  ind = '  '*depth
  for line in lines:
    nl = '' if line.endswith('\n') else '\n'
    yield f'{ind}{line}{nl}'


def iter_str(iterable:Iterable[str]) -> Iterable[str]:
  'Return the iterable unless it is a string, in which case return the single-element tuple of the string.'
  if isinstance(iterable, str): return (iterable,)
  return iterable


def iter_excluding_str(iterable:Iterable[_T]) -> Iterator[_T]:
  '''
  Often we want to handle all iterables in a particular way, except for str.
  There are two common reasons why:
  * because str should be treated as an atom/leaf value in a nested structure;
  * because the fact that elements of a str are themselves strings,
    which makes naive type-based recursion over sequences impossible.
  '''
  if isinstance(iterable, str):
    raise TypeError('iter_excluding_str explictly treats str as non-iterable type')
  return iter(iterable) # raises TypeError for non-iterables.


def pluralize(count:int, name:str, plural=None, spec='') -> str:
  'Return a string of format "{count} {name}s", with optional custom plural form and format spec.'
  if count == 1:
    n = name
  elif plural is None:
    n = name + 's'
  else:
    n = plural
  return '{count:{spec}} {n}'.format(count=count, spec=spec, n=n)


def format_nonempty(fmt:str, string:str) -> str:
  'format `string` into `format` unless `string` is empty.'
  return '' if (string == '') else fmt.format(string)

def prepend_to_nonempty(prefix:str, string:str) -> str:
  'prepend `prefix` to `string` unless `string` is empty.'
  return '' if (string == '') else (prefix + string)

def append_to_nonempty(string:str, suffix:str) -> str:
  'append `prefix` to `string` unless `string` is empty.'
  return '' if (string == '') else (string + suffix)


_byte_count_dec_magnitudes = [
  ('B',  'byte'),
  ('kB', 'kilobyte'),
  ('MB', 'megabyte'),
  ('GB', 'gigabyte'),
  ('TB', 'terabyte'),
  ('PB', 'petabyte'),
  ('EB', 'exabyte'),
  ('ZB', 'zettabyte'),
  ('YB', 'yottabyte'),
]

def format_byte_count(count:int, prec=3, abbr=True) -> str:
  "Format a string for the given number of bytes, using the largest appropriate prefix (e.g. 'kB')"
  count = int(count)
  if count < 1000:
    if abbr: return '{:d} B'.format(count)
    else: return pluralize(count, 'byte')
  c = Decimal(count)
  for abbrev, full in _byte_count_dec_magnitudes:
    if c < 999: break
    if c < 1000: # must make sure that c will not round up.
      shift = 10**prec
      if round(c * shift) < 1000 * shift: break
    c /= 1000
  if prec == 0 and not abbr:
    return pluralize(round(c), full)
  # with precision > 0, always pluralize the full names, even if all the digits are zero.
  s = '' if abbr else 's'
  label = abbrev if abbr else full
  return f'{c:.{prec}f} {label}{s}'


def line_col_0(string:str, pos:int) -> Tuple[int, int]:
  if pos < 0 or pos > len(string): raise IndexError(pos)
  line = string.count('\n', 0, pos) # number of newlines preceeding pos.
  last_line_start = string.rfind('\n', 0, pos) + 1 # rfind returns -1 for no match; `+ 1` happens to work perfectly.
  return (line, pos - last_line_start)


def line_col_1(string:str, pos:int) -> Tuple[int, int]:
  l, c = line_col_0(string, pos)
  return (l + 1, c + 1)


def simplify_punctuation(text:str) -> str:
  text = non_ascii_hyphens_re.sub('-', text) # Replace unicode hyphens.
  text = text.replace('\u2014', '--') # Em dash.
  text = non_ascii_single_quotes.sub("'", text)
  text = non_ascii_double_quotes.sub('"', text)
  return text

non_ascii_single_quotes = re.compile(r'[‘’]')
non_ascii_double_quotes = re.compile(r'[“”]')

non_ascii_hyphens_re = re.compile(r'[\xad\u2010\u2011\u2013\u2212]+')
#^ Soft hyphen, hyphen, non-breaking hyphen, en-dash, minus-sign.

zero_width_space = '\u200b'



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import os as _os
from os import R_OK, X_OK, access as _access, getpid as _getpid, supports_effective_ids as _supports_effective_ids
from os.path import dirname as _dir_name, exists as _path_exists, isfile as _is_file, join as _path_join
from selectors import EVENT_READ, EVENT_WRITE, PollSelector as _PollSelector
from shlex import quote as sh_quote, split as sh_split
from subprocess import DEVNULL, PIPE, Popen as _Popen
from sys import stderr, stdout
from time import time as _now
from typing import IO, AnyStr, BinaryIO, Dict, Iterator, List, Optional, Sequence, Tuple, Union, cast

from .alarm import Alarm, Timeout


Cmd = Union[str, Sequence[str]]
Env = Dict[str, str]
Input = Union[None, int, str, bytes, BinaryIO] # int primarily for DEVNULL; could also be raw file descriptor?
File = Union[int, IO]
ExitOpt = Union[bool, int, str]


class NonzeroCodeExpectation:
  'Type for a special marker value NONZERO, which matches any nonzero process exit code.'
  def __repr__(self): return 'NONZERO'

NONZERO = NonzeroCodeExpectation()


TaskCodeExpectation = Union[None, int, NonzeroCodeExpectation]


def launch(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None, out: File=None, err: File=None, files: Sequence[File]=(),
 note_cmd=False, lldb=False) -> Tuple[Tuple[str, ...], _Popen, Optional[bytes]]:
  '''
  Launch a subprocess, returning the normalized command as a tuple, the subprocess.Popen object and the optional input bytes.

  The underlying Subprocess shell option is not supported
  because the rules regarding splitting strings are complex.
  User code is made clearer by just specifying the complete shell command.

  If `cmd` is a list, it is used as is. If `cmd` is a string it is split by shlex.split.

  TODO: Popen supports both text and binary files; we should too.
  TODO: support bufsize parameter.
  '''

  if note_cmd: print('cmd:', fmt_cmd(cmd), file=stderr)
  if isinstance(cmd, str):
    cmd = tuple(sh_split(cmd))
  else:
    cmd = tuple(cmd)
  input_bytes: Optional[bytes]
  f_in: Input
  if isinstance(stdin, str):
    f_in = PIPE
    input_bytes = stdin.encode('utf-8')
  elif isinstance(stdin, bytes):
    f_in = PIPE
    input_bytes = stdin
  else:
    f_in = stdin # presume None, PIPE, file, or DEVNULL.
    input_bytes = None

  fds = [f if isinstance(f, int) else f.fileno for f in files]

  # flushing std file descriptors guarantees consistent behavior between console and iotest;
  # otherwise we see that parent output appears after child output when run under iotest only.
  stderr.flush()
  stdout.flush()

  try:
    proc = _Popen(
      cmd,
      cwd=cwd,
      stdin=f_in,
      stdout=out,
      stderr=err,
      shell=False,
      env=env,
      pass_fds=fds,
      preexec_fn=(preexec_launch_lldb if lldb else None))

    return cmd, proc, input_bytes

  # _Popen may raise FileNotFoundError, PermissionError, or OSError.
  # The distinction is more confusing than helpful; therefore we handle them all as OSError.
  except OSError as e:
    cmd_path_as_invoked = cmd[0] # The path as seen by the command.
    path = cmd_path_as_invoked if cwd is None else _path_join(cwd, cmd_path_as_invoked) # The cmd relative to the parent cwd, or absolute.
    # TODO: If absolute, try to make path relative to parent cwd.
    if e.filename == cmd_path_as_invoked:
      _diagnose_launch_error(path, cmd_path_as_invoked, e) # Raises a more specific exception or else return.
    raise TaskLaunchUndiagnosedError(path) from e # Default.


def preexec_launch_lldb():
  '''
  Note: this relies on a race condition to work: the spawn is slow, which gives this process time to exec.
  This is not perfect; if the child process crashes very fast then LLDB might not attach in time.
  However if we try to sleep here then LLDB stops once the exec occurs, which is useless.
  GDB has something called `follow-fork-mode` that sounds like it would address this, but sticking with LLDB for now.
  '''
  pid_str = str(_getpid())
  lldb_cmd = ['PATH=/usr/bin', 'lldb', '--batch', '--one-line', 'continue', '--attach-pid', pid_str]
  lldb_str = ' '.join(lldb_cmd)
  script = f'tell application "Terminal" to do script "{lldb_str}"'
  _os.spawnvp(_os.P_WAIT, 'osascript', ['osascript', '-e', script])
  #^ Use spawn because subprocess is complex and preexec_fn is documented to be incompatible with threading.


def _diagnose_launch_error(path:str, cmd_path:str, e:OSError) -> None:
  if not _dir_name(cmd_path): # invoked as installed command.
    if _path_exists(path): raise TaskFileInvokedAsInstalledCommand(path) from e
    else: raise TaskInstalledCommandNotFound(cmd_path) from e

  if not _path_exists(path): raise TaskFileNotFound(path) from e
  if not _is_file(path): raise TaskNotAFile(path) from e
  if not _is_permitted(path, X_OK): raise TaskFileNotExecutable(path) from e

  bad_format = (e.strerror == 'Exec format error')
  if bad_format and not _is_permitted(path, R_OK): raise TaskFileNotReadable(path) from e # Read bit is necessary for scripts.

  if bad_format or isinstance(e, FileNotFoundError):
    # The 'file not found' exception might actually be due to mistyped shebang, confusingly.
    try: # Heuristic to diagnose bad shebang lines.
      with open(path, 'rb') as f:
        lead_bytes = f.read(256) # Realistically a shebang line should not be longer than this.
        line, newline, _ = lead_bytes.partition(b'\n')
        if line and (not newline or b'\0' in line): raise TaskFileBinaryIllFormed(path, line) from e # TODO: diagnose further?
        if not line.startswith(b'#!'): raise TaskFileHashbangMissing(path, line) from e
        raise TaskFileHashbangIllFormed(path, line) from e
    except (OSError, IOError):
      raise TaskLaunchError(path) from e # open or read failed; raise the original exception.


def communicate(proc: _Popen, input_bytes: bytes=None, timeout: int=0) -> Tuple[int, bytes, bytes]:
  '''
  Communicate with and wait for a task.
  Returns (exit_code, out_bytes, err_bytes).
  '''

  # Note: Popen provides its own timeout mechanism, based on select.
  # However, the CPython implementation of communicate() has an optimized path that is only used
  # when the timeout feature is not used.
  # The tradeoff between that implementation and this alarm-based one should be examined further.
  with Alarm(timeout=timeout, msg='process timed out after {timeout} seconds and was killed', on_signal=proc.kill):
    out_bytes, err_bytes = proc.communicate(input_bytes) # waits for process to complete.

  return proc.returncode, (b'' if out_bytes is None else out_bytes), (b'' if err_bytes is None else err_bytes)


def run_gen(cmd: Cmd, cwd: str=None, env: Env=None, stdin=None, timeout: int=0, exp: TaskCodeExpectation=0,
 as_lines=True, as_text=True, merge_err=False, exits:ExitOpt=False) -> Iterator[AnyStr]:
  send: Optional[int] = None
  recv: Optional[int] = None
  if stdin == PIPE:
    stdin, send = _os.pipe()
  recv, out = _os.pipe()
  def cleanup() -> None:
    nonlocal send, recv
    if send is not None: _os.close(send)
    if recv is not None: _os.close(recv)

  try:
    cmd, proc, _ = launch(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=out, err=(out if merge_err else None))
    if send: _os.close(stdin)
    if recv: _os.close(out)

    sel = _PollSelector()
    if send: sel.register(send, EVENT_WRITE)
    if recv: sel.register(recv, EVENT_READ)
    time_start = _now()
    send_buffer: List[bytes] = []
    recv_bytes = b''

    while recv_bytes or sel.get_map():
      time_rem: Optional[float]
      if timeout > 0:
        time_rem = timeout - (_now() - time_start)
        if time_rem <= 0:
          proc.kill()
          raise Timeout(f'process timed out after {timeout} seconds and was killed')
      else:
        time_rem = None

      recv_ready = None
      send_ready = None
      if send or recv:
        for key, flags in sel.select(time_rem):
          if key.fd is recv: recv_ready = flags
          elif key.fd is send: send_ready = flags
          else: raise Exception(f'Task: received unknown selector key: {key}')

      if recv_ready:
        b = _os.read(key.fd, 0x8000)
        if b:
          recv_bytes += b
        else:
          assert recv is not None
          sel.unregister(recv)
          _os.close(recv)
          recv = None

      yield_bytes = b''
      if as_lines:
        idx = recv_bytes.find(0x0A)
        if idx >= 0: # found newline.
          p = idx + 1 # include newline.
          yield_bytes = recv_bytes[:p]
          recv_bytes = recv_bytes[p:]
        elif recv is None: # missing final newline.
          yield_bytes = recv_bytes
          recv_bytes = b''
      else:
        yield_bytes = recv_bytes
        recv_bytes = b''

      if yield_bytes or send_ready:
        send_bytes = yield cast(AnyStr, yield_bytes.decode('utf8') if as_text else yield_bytes)
        if send_bytes is not None:
          send_buffer.append(send_bytes)

      if send_ready:
        b = b''.join(send_buffer)
        send_buffer.clear()
        assert send is not None
        try: _os.write(send, b)
        except BrokenPipeError:
          sel.unregister(send)
          _os.close(send)
          send = None

    if recv: _os.close(recv)
    if send: _os.close(send)
    time_rem = timeout - (_now() - time_start)
    code = proc.wait(timeout=(time_rem if timeout > 0 else None))
    _check_exp(cmd, exp, code, exits)
    return code # generator will raise StopIteration(code).
  except BaseException:
    proc.kill()
    cleanup()
    raise


def run(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None, out: File=None, err: File=None, timeout: int=0,
 files: Sequence[File]=(), exp: TaskCodeExpectation=0, note_cmd=False, lldb=False, exits:ExitOpt=False) -> Tuple[int, str, str]:
  '''
  Run a command and return (exit_code, std_out, std_err).
  '''
  cmd, proc, input_bytes = launch(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=out, err=err, files=files, note_cmd=note_cmd,
    lldb=lldb)
  code, out_bytes, err_bytes = communicate(proc, input_bytes, timeout)
  _check_exp(cmd, exp, code, exits)
  return code, out_bytes.decode('utf8'), err_bytes.decode('utf8')


def _check_exp(cmd: Tuple[str, ...], exp: TaskCodeExpectation, code: int, exits:ExitOpt) -> None:
  if exp is None: return
  elif isinstance(exp, NonzeroCodeExpectation):
    if code != 0: return
    exp_desc = 'expected task to fail but it exited cleanly'
  else:
    if code == exp: return
    exp_desc = 'task failed' if exp == 0 else f'expected task to exit with code {exp}, but received {code}'
  if exits != False: exit(exits if isinstance(exits, str) else f'{exp_desc}; command: `{fmt_cmd(cmd)}`')
  raise UnexpectedExit(exp_desc, cmd, exp, code)


def fmt_cmd(cmd: Sequence[str]) -> str: return ' '.join(sh_quote(word) for word in cmd)


def runCOE(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None,
 timeout: int=0, files: Sequence[File]=(), note_cmd=False, lldb=False) -> Tuple[int, str, str]:
  'Run a command and return exit code, std out, std err.'
  return run(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=PIPE, err=PIPE, timeout=timeout, files=files, exp=None,
    note_cmd=note_cmd, lldb=lldb)


def runC(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None, out: BinaryIO=None, err: BinaryIO=None,
 timeout: int=0, files: Sequence[File]=(), note_cmd=False, lldb=False) -> int:
  'Run a command and return exit code; optional out and err.'
  assert out is not PIPE
  assert err is not PIPE
  c, o, e = run(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=out, err=err, timeout=timeout, files=files, exp=None,
    note_cmd=note_cmd, lldb=lldb)
  assert e == ''
  assert o == ''
  return c


def runCO(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None, err: BinaryIO=None,
 timeout: int=0, files: Sequence[File]=(), note_cmd=False, lldb=False) -> Tuple[int, str]:
  'Run a command and return exit code, std out; optional err.'
  assert err is not PIPE
  c, o, e = run(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=PIPE, err=err, timeout=timeout, files=files, exp=None,
    note_cmd=note_cmd, lldb=lldb)
  assert e == '', repr(e)
  return c, o


def runCE(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None, out: BinaryIO=None,
 timeout: int=0, files: Sequence[File]=(), note_cmd=False, lldb=False) -> Tuple[int, str]:
  'Run a command and return exit code, std err; optional out.'
  assert out is not PIPE
  c, o, e = run(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=out, err=PIPE, timeout=timeout, files=files, exp=None,
    note_cmd=note_cmd, lldb=lldb)
  assert o == ''
  return c, e


def runOE(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None,
 timeout: int=0, files: Sequence[File]=(), exp: TaskCodeExpectation=0, note_cmd=False, lldb=False, exits:ExitOpt=False) \
 -> Tuple[str, str]:
  'Run a command and return (stdout, stderr) as strings; optional code expectation `exp`.'
  c, o, e = run(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=PIPE, err=PIPE,
    timeout=timeout, files=files, exp=exp, note_cmd=note_cmd, lldb=lldb, exits=exits)
  return o, e


def runO(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None, err: BinaryIO=None,
 timeout: int=0, files: Sequence[File]=(), exp: TaskCodeExpectation=0, note_cmd=False, lldb=False, exits:ExitOpt=False) -> str:
  'Run a command and return stdout as a string; optional err and code expectation `exp`.'
  assert err is not PIPE
  c, o, e = run(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=PIPE, err=err,
    timeout=timeout, files=files, exp=exp, note_cmd=note_cmd, lldb=lldb, exits=exits)
  assert e == ''
  return o


def runE(cmd: Cmd, cwd: str=None, env: Env=None, stdin: Input=None, out: BinaryIO=None,
 timeout: int=0, files: Sequence[File]=(), exp: TaskCodeExpectation=0, note_cmd=False, lldb=False, exits:ExitOpt=False) -> str:
  'Run a command and return stderr as a string; optional out and code expectation `exp`.'
  assert out is not PIPE
  c, o, e = run(cmd=cmd, cwd=cwd, env=env, stdin=stdin, out=out, err=PIPE,
    timeout=timeout, files=files, exp=exp, note_cmd=note_cmd, lldb=lldb, exits=exits)
  assert o ==  ''
  return e


def _is_permitted(path: str, mode: int) -> bool:
  return _access(path, mode, effective_ids=(_access in _supports_effective_ids))


class ExecutableNotFound(ValueError): pass

# Exceptions.


class UnexpectedExit(Exception):
  'Exception indicating that a subprocess exit code did not match the code expectation.'
  def __init__(self, msg: str, cmd: Tuple[str, ...], exp: TaskCodeExpectation, act: int) -> None:
    super().__init__(msg)
    self.cmd = cmd
    self.exp = exp
    self.act = act


class TaskLaunchError(Exception):
  '''
  Exception indicating that `task.launch` failed.
  `launch` attempts to diagnose failures and raises a subclass of TaskLaunchError from the original.
  '''

  path: str

  @property
  def diagnosis(self) -> str:
    return f'task launch failed.'


class TaskLaunchUndiagnosedError(TaskLaunchError):
  '''
  Undiagnosed launch error.
  '''

  def __init__(self, path: str) -> None:
    super().__init__(path)
    self.path = path

  @property
  def diagnosis(self) -> str:
    return f'task launch failed (undiagnosed).'


class TaskFileBinaryIllFormed(TaskLaunchError):

  def __init__(self, path: str, first_line: bytes) -> None:
    super().__init__(path, first_line)
    self.path = path
    self.first_line = first_line

  @property
  def diagnosis(self) -> str:
    return f'file appears to be a binary; perhaps it is corrupt or the wrong format? first line: {_try_decode_repr(self.first_line)}'


class TaskFileInvokedAsInstalledCommand(TaskLaunchError):

  def __init__(self, path: str) -> None:
    super().__init__(path)
    self.path = path

  @property
  def diagnosis(self) -> str: return 'file exists but invocation is missing a leading `./`.'


class TaskFileHashbangMissing(TaskLaunchError):

  def __init__(self, path: str, first_line: bytes) -> None:
    super().__init__(path, first_line)
    self.path = path
    self.first_line = first_line

  @property
  def diagnosis(self) -> str:
    return f'script is missing shebang line (`#!...`); first line: {_try_decode_repr(self.first_line)}'


class TaskFileHashbangIllFormed(TaskLaunchError):

  def __init__(self, path: str, first_line: bytes) -> None:
    super().__init__(path, first_line)
    self.path = path
    self.first_line = first_line

  @property
  def diagnosis(self) -> str:
    return f'script shebang line may be ill-formed; first line: {_try_decode_repr(self.first_line)}'


class TaskFileNotExecutable(TaskLaunchError):

  def __init__(self, path: str) -> None:
    super().__init__(path)
    self.path = path

  @property
  def diagnosis(self) -> str: return 'file is not executable.'


class TaskFileNotFound(TaskLaunchError):

  def __init__(self, path: str) -> None:
    super().__init__(path)
    self.path = path

  @property
  def diagnosis(self) -> str: return 'file was not found.'


class TaskFileNotReadable(TaskLaunchError):

  def __init__(self, path: str) -> None:
    super().__init__(path)
    self.path = path

  @property
  def diagnosis(self) -> str: return 'file is not readable.'


class TaskInstalledCommandNotFound(TaskLaunchError):

  def __init__(self, path: str) -> None:
    super().__init__(path)
    self.path = path

  @property
  def diagnosis(self) -> str: return f'installed executable was not found: `{self.path}`'


class TaskNotAFile(TaskLaunchError):

  def __init__(self, path: str) -> None:
    super().__init__(path)
    self.path = path

  @property
  def diagnosis(self) -> str: return 'invocation path refers to a non-file.'


def _try_decode_repr(b: bytes) -> str:
  try: return repr(b.decode())
  except UnicodeError: return repr(b)



# Derived from CPython 3.7 Lib/http/server.py.
# Changes dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import mimetypes
import os
import sys
import time
import urllib.parse
from email.utils import formatdate as format_email_date
from html import escape as html_escape
from http import HTTPStatus
from http.client import HTTPException, HTTPMessage, LineTooLong, parse_headers  # type: ignore
from io import BytesIO
from os import fstat as os_fstat
from posixpath import splitext
from shutil import copyfileobj
from socket import getfqdn as get_fully_qualified_domain_name, timeout as SocketTimeout
from socketserver import StreamRequestHandler, TCPServer
from sys import exc_info
from traceback import print_exception
from typing import Any, BinaryIO, List, Optional, Tuple, Type
from urllib.parse import unquote as url_unquote, urlsplit as url_split, urlunsplit as url_join

from ..fs import is_dir, norm_path, path_exists, path_join
from ..io import errL, errSL


__version__ = '0.1'

'''
- log requests even later (to capture byte count)
- log user-agent header and other interesting goodies
- send error log to separate file
'''

# Here's a quote from the NCSA httpd docs about log file format.
#
# The logfile format is as follows. Each line consists of:
#
# host rfc931 authuser [DD/Mon/YYYY:hh:mm:ss] "request" ddd bbbb
#
# host: Either the DNS name or the IP number of the remote client
# rfc931: Any information returned by identd for this person, - otherwise.
# authuser: If user sent a userid for authentication, the user name, - otherwise.
# DD: Day
# Mon: Month (calendar name)
# YYYY: Year
# hh: hour (24-hour format, the machine's timezone)
# mm: minutes
# ss: seconds
# request: The first line of the HTTP request as sent by the client.
# ddd: the status code returned by the server, - if not available.
# bbbb: the total number of bytes sent, *not including the HTTP/1.0 header*, - if not available.
# | You can determine the name of the file accessed through request.
#
# (Actually, the latter is only true if you know the server configuration
# at the time the request was made!)



__version__ = '1'

DEFAULT_ERROR_MESSAGE = '''\
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8" />
  <title>Error: {code}</title>
</head>
<body>
  <h1>Error: {code} - {label}</p>
  <p>{message}</p>
</body>
</html>
'''

DEFAULT_ERROR_CONTENT_TYPE = "text/html;charset=utf-8"


class HTTPServer(TCPServer):

  allow_reuse_address = True # Unclear what this implies.

  unrecoverable_exceptions: Tuple[Type, ...] = (
    AttributeError,
    ImportError,
    MemoryError,
    NameError,
    SyntaxError,
  )

  def server_bind(self) -> None:
    '''Override TCP.server_bind to store the server name.'''
    super().server_bind()
    host, port = self.server_address[:2]
    self.name = get_fully_qualified_domain_name(host)
    self.port = port

  def handle_error(self, request, client_address) -> None:
    '''
    Override BaseServer.handle_error to fail fast for unrecoverable errors.
    '''
    errL()
    errSL('Exception while processing request from client:', client_address)
    e_type, exc, traceback = exc_info()
    if isinstance(exc, self.unrecoverable_exceptions):
      raise Exception('unrecoverable error') from exc
    else:
      print_exception(e_type, exc, traceback)
      errL('-'*40)


class HTTPRequestHandler(StreamRequestHandler):

  sys_version = 'Python/{}.{}.{}'.format(*sys.version_info[:3])

  # The format is multiple whitespace-separated strings, where each string is of the form name[/version].
  server_version = f"pithy.http.Server/{__version__}"

  error_message_format = DEFAULT_ERROR_MESSAGE
  error_content_type = DEFAULT_ERROR_CONTENT_TYPE


  weekdayname = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

  monthname = [None, 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

  protocol_version = "HTTP/1.1"

  MessageClass = HTTPMessage

  # hack to maintain backwards compatibility
  responses = { v: (v.phrase, v.description) for v in HTTPStatus.__members__.values() }

  # The default request version.
  # This only affects responses up until the point where the request line is parsed,
  # so it mainly decides what the client gets back when sending a malformed request line.
  # Most web servers default to HTTP 0.9, i.e. don't send a status line.
  default_request_version = protocol_version


  if not mimetypes.inited: mimetypes.init()
  ext_mime_types = mimetypes.types_map.copy()
  ext_mime_types.update({
    '': 'text/plain', # Default.
  })


  def __init__(self, *args, directory:str=None, **kwargs) -> None:
    self.command:str = '' # Set in case of error on the first line.
    self.request_version = self.default_request_version
    self.close_connection = True
    self.requestline = ''
    self.directory = directory or os.getcwd()
    self.local_path: Optional[str] = None
    self.prevent_client_caching = True
    super().__init__(*args, **kwargs)


  def handle(self) -> None:
    '''Handle multiple requests if necessary.'''
    self.close_connection = True

    self.handle_one_request()
    while not self.close_connection:
      self.handle_one_request()


  def handle_one_request(self) -> None:
    '''
    Handle a single HTTP request.
    '''
    try:
      self.raw_requestline = self.rfile.readline(65537)
      if len(self.raw_requestline) > 65536:
        self.requestline = ''
        self.request_version = ''
        self.command = ''
        self.send_error(HTTPStatus.REQUEST_URI_TOO_LONG)
        return
      if not self.raw_requestline:
        self.close_connection = True
        return
      if not self.parse_request():
        # An error code has been sent, just exit
        return
      mname = 'do_' + self.command
      if not hasattr(self, mname):
        self.send_error(
          HTTPStatus.NOT_IMPLEMENTED,
          "Unsupported method (%r)" % self.command)
        return
      method = getattr(self, mname)
      method()
      self.wfile.flush() #actually send the response if not already done.
    except SocketTimeout as e:
      #a read or a write timed out.  Discard this connection
      self.log_error("Request timed out: %r", e)
      self.close_connection = True
      return


  def parse_request(self) -> bool:
    version = self.request_version
    self.requestline = requestline = str(self.raw_requestline, 'latin-1').rstrip('\r\n')
    words = requestline.split()
    if not words: return False
    if len(words) < 3:
      self.send_error(HTTPStatus.BAD_REQUEST, "Bad request syntax (%r)" % requestline)
      return False
    version = words[-1]
    try:
      if not version.startswith('HTTP/'): raise ValueError
      base_version_number = version.split('/', 1)[1]
      version_numbers = base_version_number.split(".")
      # RFC 2145 section 3.1 says:
      # * there can be only one ".";
      # * major and minor numbers MUST be treated as separate integers;
      # * HTTP/2.4 is a lower version than HTTP/2.13, which in turn is lower than HTTP/12.3;
      # * Leading zeros MUST be ignored by recipients.
      if len(version_numbers) != 2: raise ValueError
      version_number = (int(version_numbers[0]), int(version_numbers[1]))
    except (ValueError, IndexError):
      self.send_error(HTTPStatus.BAD_REQUEST, "Bad request version (%r)" % version)
      return False
    if version_number >= (1, 1) and self.protocol_version >= "HTTP/1.1": # TODO: the lexicographical comparison here is incorrect.
      self.close_connection = False
    if version_number >= (2, 0):
      self.send_error(HTTPStatus.HTTP_VERSION_NOT_SUPPORTED, "Invalid HTTP version (%s)" % base_version_number)
      return False
    self.request_version = version

    self.command, self.path = words[:2]

    # Examine the headers and look for a Connection directive.
    try:
      self.headers = parse_headers(self.rfile, _class=self.MessageClass)
    except LineTooLong as err:
      self.send_error(HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE, "Line too long", str(err))
      return False
    except HTTPException as err:
      self.send_error(HTTPStatus.REQUEST_HEADER_FIELDS_TOO_LARGE, "Too many headers", str(err))
      return False

    conntype = self.headers.get('Connection', "")
    if conntype.lower() == 'close':
      self.close_connection = True
    elif (conntype.lower() == 'keep-alive' and self.protocol_version >= "HTTP/1.1"): # TODO: lexicographical compare.
      self.close_connection = False

    expect = self.headers.get('Expect', "")
    if (expect.lower() == "100-continue" and self.protocol_version >= "HTTP/1.1" and self.request_version >= "HTTP/1.1"): # TODO: lexicographical compare.
      if not self.handle_expect_100(): return False

    # Compute local_path.
    p = self.path
    # abandon query and fragment parameters.
    p = p.partition('?')[0]
    p = p.partition('#')[0]
    has_trailing_slash = p.rstrip().endswith('/') # remember explicit trailing slash.
    p = url_unquote(p)
    p = norm_path(p)
    if '..' in p: return False
    assert p.startswith('/')
    p = p[1:] # Remove leading slash. TODO: path_join should not use os.path.join, which behaves dangerously for absolute paths.
    p = norm_path(path_join(self.directory, p))
    if has_trailing_slash: p += '/'
    self.local_path = p
    return True


  def handle_expect_100(self) -> bool:
    '''Decide what to do with an "Expect: 100-continue" header.

    If the client is expecting a 100 Continue response,
    we must respond with either a 100 Continue or a final response before waiting for the request body.
    The default is to always respond with a 100 Continue.
    You can behave differently (for example, reject unauthorized requests) by overriding this method.

    This method should either return True (possibly after sending a 100 Continue response)
    or send an error response and return False.
    '''
    self.send_response_only(HTTPStatus.CONTINUE)
    self.end_headers()
    return True


  def send_error(self, code:HTTPStatus, message:str=None, explain:str=None) -> None:
    '''Send and log an error reply.

    Arguments are
    * code:  an HTTP error code
           3 digits
    * message: a simple optional 1 line reason phrase.
           *( HTAB / SP / VCHAR / %x80-FF )
           defaults to short entry matching the response code
    * explain: a detailed message defaults to the long entry
           matching the response code.

    This sends an error response (so it must be called before any
    output has been generated), logs the error, and finally sends
    a piece of HTML explaining the error to the user.

    '''

    try:
      shortmsg, longmsg = self.responses[code]
    except KeyError:
      shortmsg, longmsg = '???', '???'
    if message is None:
      message = shortmsg
    if explain is None:
      explain = longmsg
    self.log_error("code %d, message %s", code, message)
    self.send_response(code, message)
    self.send_header('Connection', 'close')

    # Message body is omitted for cases described in:
    #  - RFC7230: 3.3. 1xx, 204(No Content), 304(Not Modified)
    #  - RFC7231: 6.3.6. 205(Reset Content)
    body = None
    if (code >= 200 and
      code not in (HTTPStatus.NO_CONTENT,
             HTTPStatus.RESET_CONTENT,
             HTTPStatus.NOT_MODIFIED)):
      # HTML encode to prevent Cross Site Scripting attacks
      # (see bug #1100201)
      content = (self.error_message_format % {
        'code': code,
        'message': html_escape(message, quote=False),
        'explain': html_escape(explain, quote=False)
      })
      body = content.encode('UTF-8', 'replace')
      self.send_header("Content-Type", self.error_content_type)
      self.send_header('Content-Length', int(len(body)))
    self.end_headers()

    if self.command != 'HEAD' and body:
      self.wfile.write(body)


  def send_response(self, code:HTTPStatus, message:str=None) -> None:
    '''
    Add the response header to the headers buffer and log the response code.

    Also send two standard headers with the server software
    version and the current date.

    '''
    self.log_request(code)
    self.send_response_only(code, message)
    self.send_header('Server', self.version_string())
    self.send_header('Date', self.format_header_date())
    if self.prevent_client_caching:
      self.send_header('Cache-Control', 'no-cache, no-store, must-revalidate')
      self.send_header('Pragma', 'no-cache')
      self.send_header('Expires', '0')


  def send_response_only(self, code:HTTPStatus, message:str=None) -> None:
    '''Send the response header only.'''
    if self.request_version != 'HTTP/0.9':
      if message is None:
        if code in self.responses:
          message = self.responses[code][0]
        else:
          message = ''
      if not hasattr(self, '_headers_buffer'):
        self._headers_buffer:List[bytes] = []
      self._headers_buffer.append(("%s %d %s\r\n" % (self.protocol_version, code, message)).encode('latin-1', 'strict'))


  def send_header(self, keyword:str, value:Any) -> None:
    '''Send a MIME header to the headers buffer.'''
    if self.request_version != 'HTTP/0.9':
      if not hasattr(self, '_headers_buffer'):
        self._headers_buffer = []
      self._headers_buffer.append(("%s: %s\r\n" % (keyword, value)).encode('latin-1', 'strict'))

    if keyword.lower() == 'connection':
      if value.lower() == 'close':
        self.close_connection = True
      elif value.lower() == 'keep-alive':
        self.close_connection = False


  def end_headers(self) -> None:
    '''Send the blank line ending the MIME headers.'''
    if self.request_version != 'HTTP/0.9':
      self._headers_buffer.append(b"\r\n")
      self.flush_headers()


  def flush_headers(self) -> None:
    if hasattr(self, '_headers_buffer'):
      self.wfile.write(b"".join(self._headers_buffer))
      self._headers_buffer = []


  def log_error(self, format:str, *args:Any) -> None:
    '''
    Log an error.

    This is called when a request cannot be fulfilled. By default it passes the message on to log_message().
    Arguments are the same as for log_message().
    XXX This should go to the separate error log.
    '''
    self.log_message(format, *args)


  def version_string(self) -> str:
    '''Return the server software version string.'''
    return self.server_version + ' ' + self.sys_version


  def address_string(self) -> str:
    '''Return the client address.'''
    return self.client_address[0] # type: ignore


  def do_GET(self) -> None:
    '''Serve a GET request.'''
    f = self.send_head()
    if f:
      try: copyfileobj(f, self.wfile)
      finally: f.close()


  def do_HEAD(self) -> None:
    '''Serve a HEAD request.'''
    f = self.send_head()
    if f:
      f.close()


  def send_head(self) -> Optional[BinaryIO]:
    if self.path == '/favicon.ico': # TODO: send actual favicon if it exists.
      self.send_response(HTTPStatus.OK)
      self.send_header('Content-type', 'image/x-icon')
      self.send_header('Content-Length', 0)
      self.end_headers()
      return None

    if self.local_path is None:
      self.send_error(HTTPStatus.UNAUTHORIZED, "Path refers to parent directory")
      return None

    if is_dir(self.local_path, follow=True):
      if not self.local_path.endswith('/'): # redirect browser to path with slash (what apache does).
        self.send_response(HTTPStatus.MOVED_PERMANENTLY)
        parts = list(url_split(self.path))
        parts[2] += '/'
        new_url = url_join(parts)
        self.send_header("Location", new_url)
        self.end_headers()
        return None
      for index in ("index.html", "index.htm"):
        index = path_join(self.local_path, index)
        if path_exists(index, follow=False):
          self.local_path = index
          break
      else:
        return self.list_directory(self.local_path)

    ctype = self.guess_file_type(self.local_path)
    f: BinaryIO
    try:
      f = open(self.local_path, 'rb')
    except OSError:
      self.send_error(HTTPStatus.NOT_FOUND,
        message=f'File not found: {self.local_path}',
        explain=f'URI path: {self.path}')
      return None
    try:
      f_stat = os_fstat(f.fileno())
      self.send_response(HTTPStatus.OK)
      self.send_header("Content-type", ctype)
      self.send_header("Content-Length", f_stat.st_size)
      self.send_header("Last-Modified", self.format_header_date(f_stat.st_mtime))
      self.end_headers()
      return f
    except:
      f.close()
      raise


  def list_directory(self, path:str) -> Optional[BytesIO]:
    '''
    Helper to produce a directory listing (absent index.html).

    Return value is either a file object, or None (indicating an error).
    In either case, the headers are sent, making the interface the same as for send_head().
    '''
    try:
      list = os.listdir(path)
    except OSError:
      self.send_error(
        HTTPStatus.NOT_FOUND,
        "No permission to list directory")
      return None
    list.sort(key=lambda a: a.lower())
    r = []
    try:
      displaypath = urllib.parse.unquote(self.path,
                         errors='surrogatepass')
    except UnicodeDecodeError:
      displaypath = urllib.parse.unquote(path)
    displaypath = html_escape(displaypath, quote=False)
    enc = sys.getfilesystemencoding()
    title = 'Directory listing for %s' % displaypath
    r.append('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">')
    r.append('<html>\n<head>')
    r.append('<meta http-equiv="Content-Type" content="text/html; charset=%s">' % enc)
    r.append('<title>%s</title>\n</head>' % title)
    r.append('<body>\n<h1>%s</h1>' % title)
    r.append('<hr>\n<ul>')
    for name in list:
      fullname = os.path.join(path, name)
      displayname = linkname = name
      # Append / for directories or @ for symbolic links
      if os.path.isdir(fullname):
        displayname = name + "/"
        linkname = name + "/"
      if os.path.islink(fullname):
        displayname = name + "@"
        # Note: a link to a directory displays with @ and links with /
      r.append('<li><a href="%s">%s</a></li>' % (urllib.parse.quote(linkname, errors='surrogatepass'), html_escape(displayname, quote=False)))
    r.append('</ul>\n<hr>\n</body>\n</html>\n')
    encoded = '\n'.join(r).encode(enc, 'surrogateescape')
    f = BytesIO()
    f.write(encoded)
    f.seek(0)
    self.send_response(HTTPStatus.OK)
    self.send_header("Content-type", "text/html; charset=%s" % enc)
    self.send_header("Content-Length", str(len(encoded)))
    self.end_headers()
    return f


  def guess_file_type(self, path:str) -> str:
    base, ext = splitext(path)
    ext = ext.lower()
    try: return self.ext_mime_types[ext]
    except KeyError: return self.ext_mime_types['']


  def log_message(self, format:str, *args:Any) -> None:
    'Base logging function called by all others. Overridden to alter formatting.'
    errL(f'{self.format_log_date()}: {self.address_string()} - {format%args}')


  def log_request(self, code='-', size='-') -> None:
    'Log an accepted request; called by send_response().'
    assert isinstance(code, HTTPStatus)
    self.log_message('%s %s "%s": %s', code.value, code.phrase, self.requestline, self.local_path)


  def format_log_date(self, timestamp:float=None) -> str:
    'Format the current time for logging.'
    if timestamp is None: timestamp = time.time()
    ts = time.localtime(timestamp)
    y, m, d, hh, mm, ss, wd, yd = ts[:8]
    return f'{y:04}-{m:02}-{d:02} {hh:02}:{mm:02}:{ss:02}.{timestamp%1:.03f}'


  def format_header_date(self, timestamp:float=None):
    'Format `timestamp` or now for an HTTP header value.'
    return format_email_date(time.time() if timestamp is None else timestamp, usegmt=True)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Any, Callable, Iterable, NamedTuple, Tuple, Type, TypeVar


class lazy_property(object):
  'Lazy property decorator.'

  def __init__(self, acc_fn:Callable) -> None:
    self.acc_fn = acc_fn

  def __get__(self, obj:Any, cls:Type) -> Any:
    val = self.acc_fn(obj)
    setattr(obj, self.acc_fn.__name__, val)
    return val


_T = TypeVar('_T')

def once(fn:Callable[[],_T]) -> _T: return fn()


def memoize(_fn:Callable=None, sentinel:Any=Ellipsis) -> Callable:
  '''
  recursive function memoization decorator.
  results will be memoized by a key that is the tuple of all arguments.
  the sentinel is inserted into the dictionary before the call.
  thus, if the function recurses with identical arguments the sentinel will be returned to the inner calls.
  '''

  def _memoize(fn:Callable) -> Callable:

    class MemoDict(dict):
      def __repr__(self) -> str: return f'@memoize({sentinel}){fn}'
      def __call__(self, *args:Any) -> Any: return self[args]
      def __missing__(self, args:Any) -> Any:
        self[args] = sentinel
        res = fn(*args)
        self[args] = res
        return res

    return MemoDict()

  if _fn is None: # called parens.
    return _memoize
  else: # called without parens.
    return _memoize(_fn)


def nt_items(nt:NamedTuple) -> Iterable[Tuple[str,Any]]:
  'Return an iterable that returns the (name, value) pairs of a NamedTuple.'
  return zip(nt._fields, nt)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import json as _json
import re
from dataclasses import fields, is_dataclass
from json.decoder import JSONDecodeError
from sys import stderr, stdout
from typing import IO, AbstractSet, Any, Callable, Dict, FrozenSet, Iterable, List, Optional, Sequence, TextIO, Tuple, Union

from .encode import EncodeObj, all_slots, encode_obj


JsonAny = Any # TODO: remove this once recursive types work.
JsonList = List[JsonAny]
JsonDict = Dict[str, JsonAny]
Json = Union[None, int, float, str, bool, JsonList, JsonDict]

JsonText = Union[str,bytes,bytearray]

class JSONEmptyDocument(JSONDecodeError): pass

ObjDecodeFn = Callable[[Dict],Any]
ObjDecodeHook = Union[type, Tuple[AbstractSet[str],Union[type,ObjDecodeFn]]]
ObjDecodeHooks = Sequence[ObjDecodeHook]

_Seps = Optional[Tuple[str,str]]


def render_json(item:Any, default:EncodeObj=encode_obj, sort=True, indent:Optional[int]=2, separators:_Seps=None, **kwargs) -> str:
  'Render `item` as a json string.'
  if not separators:
    separators = (',', ': ') if indent else (',', ':')
  return _json.dumps(item, indent=indent, default=default, sort_keys=sort, separators=separators, **kwargs)


def write_json(file:TextIO, *items:Any, default:EncodeObj=encode_obj, sort=True, indent:Optional[int]=2, separators:_Seps=None, end='\n', flush=False, **kwargs) -> None:
  'Write each item in `items` as json to file.'
  if not separators:
    separators = (',', ': ') if indent else (',', ':')
  for item in items:
    _json.dump(item, file, indent=indent, default=default, sort_keys=sort, separators=separators, **kwargs)
    if end: file.write(end)
    if flush: file.flush()


def err_json(*items:Any, default:EncodeObj=encode_obj, sort=True, indent:Optional[int]=2, separators:_Seps=None, end='\n', flush=False, **kwargs) -> None:
  'Write items as json to std err.'
  write_json(stderr, *items, default=default, sort=sort, indent=indent, separators=separators, end=end, flush=flush, **kwargs)


def out_json(*items:Any, default:EncodeObj=encode_obj, sort=True, indent:Optional[int]=2, separators:_Seps=None, end='\n', flush=False, **kwargs) -> None:
  write_json(stdout, *items, default=default, sort=sort, indent=indent, separators=separators, end=end, flush=flush, **kwargs)


def write_jsonl(file:TextIO, *items:Any, default:EncodeObj=encode_obj, sort=True, separators:_Seps=None, flush=False, **kwargs) -> None:
  'Write each item in `items` as jsonl to file.'
  if separators:
    assert '\n' not in separators[0]
    assert '\n' not in separators[1]
  else:
    separators = (',', ':')
  for item in items:
    _json.dump(item, file, indent=None, default=default, sort_keys=sort, separators=separators, **kwargs)
    file.write('\n')
    if flush: file.flush()


def err_jsonl(*items:Any, default:EncodeObj=encode_obj, sort=True, separators:_Seps=None, flush=False, **kwargs) -> None:
  'Write items as jsonl to std err.'
  write_jsonl(stderr, *items, default=default, sort=sort, separators=separators, flush=flush, **kwargs)


def out_jsonl(*items:Any, default:EncodeObj=encode_obj, sort=True, separators:_Seps=None, flush=False, **kwargs) -> None:
  'Write items as jsonl to std out.'
  write_jsonl(stdout, *items, default=default, sort=sort, separators=separators, flush=flush, **kwargs)


# input.


def parse_json(text:JsonText, hook:ObjDecodeFn=None, hooks:ObjDecodeHooks=()) -> Any:
  '''
  Parse json from `text`.
  If `object_hook` is None and `types` is a non-empty sequence,
  then a hook is created that transforms JSON objects into matching record types,
  based on field name sets.
  The sets of field names must be distinct for all provided record types.
  '''
  return _json.loads(text, object_hook=_mk_hook(hook, hooks))


def load_json(file:IO, hook:ObjDecodeFn=None, hooks:ObjDecodeHooks=()) -> Any:
  '''
  Read json from `file`.
  If `object_hook` is None and `types` is a non-empty sequence,
  then a hook is created that transforms JSON objects into matching record types,
  based on field name sets.
  The sets of field names must be distinct for all provided record types.
  '''
  try:
    return _json.load(file, object_hook=_mk_hook(hook, hooks))
  except JSONDecodeError as e:
    if e.pos == 0 and e.msg == 'Expecting value':
      raise JSONEmptyDocument(msg=e.msg, doc=e.doc, pos=e.pos) from e
    else:
      raise


def parse_jsonl(text:JsonText, hook:ObjDecodeFn=None, hooks:ObjDecodeHooks=()) -> Iterable[Any]:
  hook = _mk_hook(hook, hooks)
  return (_json.loads(line, object_hook=hook) for line in text.splitlines())


def load_jsonl(stream:Iterable[JsonText], hook:ObjDecodeFn=None, hooks:ObjDecodeHooks=()) -> Iterable[Any]:
  hook = _mk_hook(hook, hooks)
  return (_json.loads(line, object_hook=hook) for line in stream)


def parse_jsons(string:str, hook:ObjDecodeFn=None, hooks:ObjDecodeHooks=()) -> Iterable[Any]:
  '''
  Parse multiple json objects from `string`.
  If `types` is a non-empty sequence,
  then an object hook is passed to the decoder transforms JSON objects into matching dataclass types.
  '''
  hook = _mk_hook(hook, hooks)
  decoder = _json.JSONDecoder(object_hook=hook)
  m = _ws_re.match(string, 0)
  assert m is not None
  idx = m.end() # must consume leading whitespace for the decoder.

  while idx < len(string):
    obj, end = decoder.raw_decode(string, idx)
    yield obj
    m = _ws_re.match(string, end)
    assert m is not None
    idx = m.end()


def load_jsons(file:TextIO, hook:ObjDecodeFn=None, hooks:ObjDecodeHooks=()) -> Iterable[Any]:
  # TODO: it seems like we ought to be able to stream the file into the parser,
  # but JSONDecoder requires the entire string for a single JSON segment.
  # Therefore in order to stream we would need to read into a buffer,
  # count nesting tokens (accounting for strings and escaped characters inside of them),
  # identify object boundaries and create substrings to pass to the decoder.
  # For now just read the whole thing at once.
  return parse_jsons(file.read(), hook=hook, hooks=hooks)


def _mk_hook(hook:Optional[ObjDecodeFn], hooks:ObjDecodeHooks) -> Optional[Callable[[Dict[Any,Any]],Any]]:
  '''
  Provide a hook function to deserialize JSON into the provided types.
  `types` is a sequence of type objects, each of which must be a dataclass or NamedTuple.
  '''
  if not hooks: return hook
  dflt_hook:ObjDecodeFn = lambda d: d if hook is None else hook

  type_map:Dict[FrozenSet[str],Any] = {}
  for h in hooks:

    fn:ObjDecodeFn
    if isinstance(h, type):
      keys_raw = _hook_type_keys(h)
      fn = _hook_type_fn(h)
    else:
      try: keys_raw, fn_raw = h # Explicit pair.
      except Exception as e:
        raise ValueError(f'malformed decoder hook; expected (keys, constructor) pair; received: {h!r}') from e
      fn = _hook_type_fn(fn_raw) if isinstance(fn_raw, type) else fn_raw
      if isinstance(keys_raw, str): raise TypeError(keys_raw) # type: ignore

    keys = frozenset(keys_raw)
    if keys in type_map:
      raise ValueError(f'conflicting type hooks for key seyt {keys}:\n  {type_map[keys]}\n  {fn}')
    type_map[keys] = fn

  def types_object_hook(d:Dict) -> Any:
    keys = frozenset(d.keys())
    try: record_type = type_map[keys]
    except KeyError: return dflt_hook(d)
    return record_type(**d)

  return types_object_hook


def _hook_type_keys(t:type) -> AbstractSet[str]:
  if is_dataclass(t): return frozenset(f.name for f in fields(t))

  try: return frozenset(t._fields) # type: ignore # NamedTuple.
  except AttributeError: pass

  slots = all_slots(t)
  if slots: return frozenset(slots)
  else: raise TypeError(f'type must be either a dataclass, namedtuple, or define `__slots__`: {type}')


def _hook_type_fn(t:type) -> ObjDecodeFn:
  try: return t.from_json # type: ignore
  except AttributeError: return t


_ws_re = re.compile(r'[ \t\n\r]*') # identical to json.decoder.WHITESPACE.



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Print indented object descriptions.
like `pprint` but streaming, and with a more compact, minimal style.
'''


import re
from dataclasses import fields as _dc_fields, is_dataclass
from sys import stderr, stdout
from typing import Any, Iterable, Iterator, List, Mapping, NamedTuple, Set, TextIO, Tuple, Union

from .iterable import known_leaf_types


# Special handling for annoying types that show up in Python scopes.
_Printer = type(copyright)
_Quitter = type(quit) # Also the type of `exit`.
_BadRepr = (_Printer, _Quitter)


def writeD(file:TextIO, *labels_and_obj:Any, indent='', exact=False) -> None:
  'Write a description to a file.'
  obj = labels_and_obj[-1]
  labels = labels_and_obj[:-1]
  if labels: print(*labels, end=': ', file=file)
  for part in gen_desc(obj, indent=indent, exact=exact):
    print(part, end='', file=file)
  print(file=file)


def errD(*labels_and_obj:Any, indent='', exact=False) -> None:
  'Write a description to `stderr`.'
  writeD(stderr, *labels_and_obj, indent=indent, exact=exact)


def outD(*labels_and_obj:Any, indent='', exact=False) -> None:
  'Write a description to `stdout`.'
  writeD(stdout, *labels_and_obj, indent=indent, exact=exact)


_DescEl = Union[str,'_Desc']

class _Desc(NamedTuple):
  'Temporary description object.'
  opener:str # Opening character, e.g. '('.
  closer:str # Closing character, e.g. ')'.
  it:Iterator # TODO: when mypy supports recursive types, fully annotate with `['_DescEl']`.
  buffer:List[str] # Holds leading elements obtained from `it`.

  def scan_inlineables(self, max_width:int) -> bool:
    'Determine if this description can be rendered inline.'
    width = 0
    for el in self.it:
      self.buffer.append(el)
      if not isinstance(el, str) or ',' in el: # Comma heuristic could be configurable.
        return False
      width += len(el)
      if width > max_width: return False
    return True # Inlineable.

  def children(self) -> Iterable[_DescEl]:
    yield from self.buffer
    del self.buffer[:] # Makes the single-use semantics a bit more explicit, but not perfect.
    yield from self.it


def gen_desc(obj:Any, indent:str='', exact=False) -> Iterator[str]:
  '''
  Generate description parts. Does not include final newline.
  '''
  yield from _gen_desc(_obj_desc(obj, prefix='', visited_ids=set(), simple_keys=(not exact)), indent=indent, exact=exact)


def _gen_desc(d:_DescEl, indent:str, exact:bool) -> Iterator[str]:
  '''
  Generate description parts. Does not include final newline.
  '''
  if indent: yield indent

  if isinstance(d, str):
    yield d
    return

  yield d.opener

  if d.scan_inlineables(max_width=128): # Inlineable.
    if d.buffer: # Nonempty.
      yield ' '
      comma_joiner = ', ' if exact else ' '
      yield comma_joiner.join(d.buffer)
      yield ' '

  else: # Not inlineable.
    indent1 = indent + '  '
    is_leaf = False
    is_subsequent = False
    for el in d.children():
      yield ',\n' if (exact and is_subsequent) else '\n'
      is_subsequent = True
      is_leaf = isinstance(el, str)
      if is_leaf:
        yield indent1
        yield el # type: ignore
      else:
        yield from _gen_desc(el, indent1, exact)
    if is_leaf: yield ' ' # Final space before closer.

  yield d.closer


def _obj_desc(obj:Any, prefix:str, visited_ids:Set[int], simple_keys:bool) -> _DescEl:
  '''
  Main description generator function. This dispatches out to others using various criteria.
  '''

  if isinstance(obj, known_leaf_types):
    return prefix + repr(obj)

  i = id(obj)
  if i in visited_ids:
    return f'^0x{i:x}:{type(obj).__name__}'

  # TODO: add singledispatch override here?

  if is_dataclass(obj):
    visited_ids1 = visited_ids.copy()
    visited_ids1.add(i)
    items:_Items = ((f.name, getattr(obj, f.name, '<?>')) for f in _dc_fields(obj))
    return _record_desc(obj, prefix, visited_ids1, items, simple_keys)

  # Most objects in a tree are leaves; we minimize tests for the leaf case by nesting the mapping test inside the iter test.
  # This has the debatable side effect of ignoring non-iterable classes that have an items() function.

  # Use an explicit test for __iter__ because iter() will return for objects without __iter__ but with __getitem__;
  # this includes typing._SpecialForm, which fails when we attempt to iterate.
  if hasattr(obj, '__iter__'):
    try: _ = iter(obj)
    except TypeError: pass
    else: # `obj` is an Iterator.
      visited_ids1 = visited_ids.copy()
      visited_ids1.add(i)
      # Attempt to distinguish between mapping and sequence types.
      try: items = iter(obj.items()) # The outer call to `iter` guards against badly formed items() functions.
      except (AttributeError, TypeError): # Treat as iterable.
        return _iterable_desc(obj, prefix, visited_ids1, iter(obj), simple_keys)
      else: # Treat as a mapping.
        return _mapping_desc(obj, prefix, visited_ids1, items, simple_keys)

  if isinstance(obj, _BadRepr):
    return f'{prefix}{type(obj).__name__}<{repr(obj)!r}>'
  return prefix + repr(obj)


def _iterable_desc(obj:Any, prefix:str, visited_ids:Set[int], it:Iterator, simple_keys:bool) -> _Desc:
  t = type(obj)
  if t is tuple:
    opener = '('
    closer = ',)' if len(obj) == 1 else ')'
  elif t is list:
    opener = '['
    closer = ']'
  elif t is set:
    it = iter(sorted(it))
    if obj:
      opener = '{'
      closer = '}'
    else:
      opener = 'set('
      closer = ')'
  else:
    if isinstance(obj, (frozenset, set)):
      it = iter(sorted(it))
    opener = t.__qualname__ + '(['
    closer = '])'

  it = (_obj_desc(el, prefix='', visited_ids=visited_ids, simple_keys=simple_keys) for el in it)
  return _Desc(opener=prefix+opener, closer=closer, it=it, buffer=[])


_Items = Iterator[Tuple[Any,Any]]


def _mapping_desc(obj:Mapping, prefix:str, visited_ids:Set[int], items:_Items, simple_keys:bool) -> _Desc:
  t = type(obj)
  if t is dict:
    opener = '{'
    closer = '}'
  else:
    opener = t.__qualname__ + '({'
    closer = '})'

  it = _gen_item_descs(visited_ids, items, simple_keys, key_sep=':')
  return _Desc(opener=prefix+opener, closer=closer, it=it, buffer=[])


def _record_desc(obj:Any, prefix:str, visited_ids:Set[int], items:_Items, simple_keys:bool) -> _Desc:
  it = _gen_item_descs(visited_ids, items, simple_keys, key_sep='=')
  return _Desc(opener=f'{prefix}{type(obj).__qualname__}(', closer=')', it=it, buffer=[])


def _gen_item_descs(visited_ids:Set[int], items:_Items, simple_keys:bool, key_sep:str) -> Iterator[_DescEl]:
  for pair in items:
    try: k, v = pair
    except (TypeError, ValueError): # Guard against a weird items() iterator.
      yield f'! {pair!r}'
      continue
    if simple_keys:
      ks = str(k)
      if not _word_re.fullmatch(ks): ks = repr(k)
    else:
      ks = repr(k)
    yield _obj_desc(v, prefix=ks+key_sep, visited_ids=visited_ids, simple_keys=simple_keys)


def repr_clean(obj:Any) -> str:
  '''
  Attempt to ensure that an object gets printed with a decent repr.
  Some third-party libraries print unconventional reprs that are confusing inside of collection descriptions.
  '''
  r = repr(obj)
  if type(obj) in _literal_repr_types or _decent_repr_re.fullmatch(r): return r
  return f'{type(obj).__name__}({r})'

_literal_repr_types = { bytes,str,list,dict,set,tuple }
_decent_repr_re = re.compile(r'[a-zA-Z][.\w]*\(.*\)')
_word_re = re.compile(r'[-\w]+')


def repr_lim(obj:Any, limit=64) -> str:
  r = repr(obj)
  if limit > 2 and len(r) > limit:
    q = r[0]
    if q in '\'"': return f'{r[:limit-2]}{q}…'
    else: return f'{r[:limit-1]}…'
  return r



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Callable, Generic, Iterable, TypeVar, Union


_T = TypeVar('_T')


class DefaultList(list, Generic[_T]):
  '''
  A subclass of `list` that adds default elements produced by a factory function
  when an out-of-bounds element is accessed.
  The factory function takes the array index as a its sole parameter.
  '''

  def __init__(self, factory: Callable[[int], _T], iterable: Iterable[_T]=(), fill_length=0) -> None:
    super().__init__(iterable)
    self.factory = factory
    for i in range(fill_length):
      self.append(factory(i))

  def __getitem__(self, index: Union[int, slice]):
    if isinstance(index, slice):
      end = 0 if index.stop is None else index.stop
    else:
      end = index
    for i in range(len(self), end):
      self.append(self.factory(i))
    return super().__getitem__(index)

  def __repr__(self) -> str:
    return '{}({}, {})'.format(type(self).__qualname__, self.factory, super().__repr__())



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from abc import abstractmethod
from typing import Any, Generic, Iterable, Iterator, Protocol, Sequence, TypeVar, Union

from .default import Default


Num = Union[int, float]

_Step = TypeVar('_Step', contravariant=True)


_RangeBound = TypeVar('_RangeBound', bound='RangeBound')

class RangeBound(Protocol, Generic[_Step]):

  @abstractmethod
  def __lt__(self, other:Any) -> bool: ...

  @abstractmethod
  def __eq__(self, other:Any) -> bool: ...

  @abstractmethod
  def __add__(self:_RangeBound, other:_Step) -> _RangeBound: ...



_Bound = TypeVar('_Bound', bound=RangeBound)


class Range(Generic[_Bound, _Step], Iterable[_Bound]):

  def __init__(self, start:_Bound, stop:Union[_Bound,Default]=Default._, *, step:_Step, length:Union[_Step,Default]=Default._, closed=False) -> None:
    self.start = start
    self.step = step
    self.closed = closed
    if isinstance(stop, Default):
      if isinstance(length, Default): raise ValueError('Range requires either `stop` or `length` argument')
      try: self.stop = start + length
      except TypeError:
        # Because builtin types like DateTime do not correctly raise NotImplemented, we try reversing the operands manually.
        try: self.stop = length + start # type: ignore
        except ValueError as e: raise ValueError(f'Range stop calculation failed: start: {start!r}; length: {length!r}') from e
    else:
      self.stop = stop

  def __iter__(self) -> Iterator[_Bound]:
    pos = self.start
    stop = self.stop
    step = self.step
    while pos < stop:
      yield pos
      try: pos += step
      except TypeError:
        # Because builtin types like DateTime do not correctly raise NotImplemented, we try reversing the operands manually.
        try: pos = step + pos # type: ignore
        except ValueError as e: raise ValueError(f'Range step addition failed: position: {pos!r}; step: {step!r}') from e
    if self.closed:
      yield stop


_setattr = object.__setattr__


class NumRange(Sequence[Num]):

  start:Num
  stop:Num
  step:Num
  closed:bool
  _len:int

  def __init__(self, start:Num, stop:Num=None, step:Num=1, *, closed=False) -> None:
    if stop is None: # Imitate `range`; `start` is actually `stop`.
      _setattr(self, 'start', 0)
      _setattr(self, 'stop', start)
    else:
      _setattr(self, 'start', start)
      _setattr(self, 'stop', stop)
    _setattr(self, 'step', step)
    _setattr(self, 'closed', closed)
    dist = max(0, self.stop - self.start)
    steps = dist / step
    extra = 1 if (closed or steps % 1) else 0
    _setattr(self, '_len', int(steps) + extra)

  def __len__(self):
    return self._len

  def __getitem__(self, i:Union[int, slice]) -> Num: # type: ignore
    if isinstance(i, int):
      if i >= 0 and i >= self._len: raise IndexError(i)
      return self.start + self.step * (i if i >= 0 else (self._len+i))
    elif isinstance(i, slice):
      raise NotImplementedError('NumRange does not yet support extended slicing')

  def __iter__(self) -> Iterator[Num]:
    return (self.start + self.step * i for i in range(self._len))

  def __repr__(self):
    return '{}({}, {}, {})'.format(type(self).__name__, self.start, self.stop, self.step)

  def __setattr__(self, name:str, val:str) -> None:
    raise AttributeError('NumRange attributes are readonly')

  def __hash__(self) -> int:
    return hash(self.start)^hash(self.stop)^hash(self.step)^int(self.closed)

  def __eq__(self, other:object) -> bool:
    return type(self) == type(other) and vars(self) == vars(other)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
ANSI Control Sequences.


ANSI Select Graphics Rendition (SGR) sequences.

RST: reset

BOLD: bold
ULINE: underline
BLINK: blink
INVERT: invert
TXT: color text
BG: color background

K: black
W: white
D: dim gray
R: red
G: green
Y: yellow
B: blue
M: magenta
C: cyan
L: light gray


Incomplete

Additional, unimplemented commands are documented below.
As these are implemented, the command chars should be added to the cs_re pattern.

CSI n A CUU – Cursor Up Moves the cursor n (default 1) cells in the given direction.
  If the cursor is already at the edge of the screen, this has no effect.

CSI n B: Cursor Down
CSI n C: Cursor Forward
CSI n D: Cursor Back
CSI n E: Moves cursor to beginning of the line n (default 1) lines down.
CSI n F: Moves cursor to beginning of the line n (default 1) lines up.

CSI n G: Moves the cursor to column n.

CSI n S: Scroll whole page up by n (default 1) lines. New lines are added at the bottom. (not ANSI.SYS)
CSI n T: Scroll whole page down by n (default 1) lines. New lines are added at the top. (not ANSI.SYS)

'''

import re as _re

from sys import stderr, stdout
from typing import Any, List

is_err_tty = stderr.isatty()
is_out_tty = stdout.isatty()

# Use these with `and` expressions to omit sgr for non-tty output, e.g. `TTY_OUT and sgr(...)`.
TTY_ERR = '!TTY_ERR' if is_err_tty else ''
TTY_OUT = '!TTY_OUT' if is_out_tty else ''

# ANSI control sequence indicator.
CSI = '\x1B['

# regex for detecting control sequences in strings.
# TODO: replace .*? wildcard with stricter character set.
ctrl_seq_re = _re.compile(r'\x1B\[.*?[hHJKlmsu]')


def ctrl_seq(c:str, *args:Any) -> str:
  'Format a control sequence string for command character `c` and arguments.'
  return f'{CSI}{";".join(str(a) for a in args)}{c}'


def strip_ctrl_seq(text: str) -> str:
  'Strip control sequences from a string.'
  return ctrl_seq_re.sub('', text)


def len_strip_ctrl_seq(s: str) -> int:
  'Calculate the length of string if control sequences were stripped.'
  l = len(s)
  for m in ctrl_seq_re.finditer(s):
    l -= m.end() - m.start()
  return l


def sgr(*seq:Any) -> str:
  'Select Graphic Rendition control sequence string.'
  return ctrl_seq('m', *seq)


# reset command strings.
RST = sgr() # Equivalent to sgr(0).
RST_ERR = (TTY_ERR and RST)
RST_OUT = (TTY_OUT and RST)

(RST_BOLD, RST_ULINE, RST_BLINK, RST_INVERT, RST_TXT, RST_BG) = (22, 24, 25, 27, 39, 49)

# effect command strings.
(BOLD, ULINE, BLINK, INVERT) = (1, 4, 5, 7)


# "Primary" colors.
# Note that black and white acronyms are suffixed with T,
# because we prefer to use true black and white from xterm-256color, defined below.

# color text: black, red, green, yellow, blue, magenta, cyan, white.
txt_primary_indices = range(30, 38)
txt_primaries = tuple(sgr(i) for i in txt_primary_indices)
TXT_KT, TXT_R, TXT_G, TXT_Y, TXT_B, TXT_M, TXT_C, TXT_WT = txt_primaries

TXT_KT_ERR, TXT_R_ERR, TXT_G_ERR, TXT_Y_ERR, TXT_B_ERR, TXT_M_ERR, TXT_C_ERR, TXT_WT_ERR = (
  (TTY_ERR and c) for c in txt_primaries)

TXT_KT_OUT, TXT_R_OUT, TXT_G_OUT, TXT_Y_OUT, TXT_B_OUT, TXT_M_OUT, TXT_C_OUT, TXT_WT_OUT = (
  (TTY_OUT and c) for c in txt_primaries)


# color background: black, red, green, yellow, blue, magenta, cyan, white.
bg_primary_indices = range(40, 48)
bg_primaries = tuple(sgr(i) for i in bg_primary_indices)
BG_KT, BG_R, BG_G, BG_Y, BG_B, BG_M, BG_C, BG_WT = bg_primaries

BG_KT_ERR, BG_R_ERR, BG_G_ERR, BG_Y_ERR, BG_B_ERR, BG_M_ERR, BG_C_ERR, BG_WT_ERR = (
  (TTY_ERR and c) for c in  bg_primaries)

BG_KT_OUT, BG_R_OUT, BG_G_OUT, BG_Y_OUT, BG_B_OUT, BG_M_OUT, BG_C_OUT, BG_WT_OUT = (
  (TTY_OUT and c) for c in  bg_primaries)

# xterm-256 sequence initiators; these should be followed by a single color index.
# both text and background can be specified in a single sgr call.
TXT = '38;5'
BG = '48;5'

# RGB6 color cube: 6x6x6, from black to white.
K = 16  # black.
W = 231 # white.
D = W + 7 # #444444.
N = W + 13 # 808080.
L = W + 18 # B2B2B2.

# Grayscale: the 24 palette values have a suggested 8 bit grayscale range of [8, 238].
middle_gray_indices = range(232, 256)

def gray26(n:int) -> int:
  assert 0 <= n < 26
  if n == 0: return K
  if n == 25: return W
  return W + n

def rgb6(r:int, g:int, b:int) -> int:
  'index RGB triples into the 256-color palette (returns 16 for black, 231 for white).'
  assert 0 <= r < 6
  assert 0 <= g < 6
  assert 0 <= b < 6
  return (((r * 6) + g) * 6) + b + 16

named_gray_indices = (K, D, N, L, W)

TXT_K, TXT_D, TXT_N, TXT_L, TXT_W = txt_grays = tuple(sgr(TXT, code) for code in named_gray_indices)

TXT_K_OUT, TXT_D_OUT, TXT_N_OUT, TXT_L_OUT, TXT_W_OUT = ((TTY_OUT and c) for c in txt_grays)
TXT_K_ERR, TXT_D_ERR, TXT_N_ERR, TXT_L_ERR, TXT_W_ERR = ((TTY_ERR and c) for c in txt_grays)

BG_K, BG_D, BG_N, BG_L, BG_W = bg_grays = tuple(sgr(BG, code) for code in named_gray_indices)

BG_K_OUT, BG_D_OUT, BG_N_OUT, BG_L_OUT, BG_W_OUT = ((TTY_OUT and c) for c in bg_grays)
BG_K_ERR, BG_D_ERR, BG_N_ERR, BG_L_ERR, BG_W_ERR = ((TTY_ERR and c) for c in bg_grays)


def cursor_pos(x:int, y:int) -> str:
  '''
  Position the cursor.
  Supposedly the 'f' suffix does the same thing.
  x and y parameters are zero based.
  '''
  return ctrl_seq('H', y + 1, x + 1)


ERASE_LINE_F, ERASE_LINE_B, ERASE_LINE = (ctrl_seq('K', i) for i in range(3))

CLEAR_SCREEN_F, CLEAR_SCREEN_B, CLEAR_SCREEN = (ctrl_seq('J', i) for i in range(3))

FILL = ERASE_LINE_F + RST # Erase-line fills the background color to the end of line.
FILL_ERR = (TTY_ERR and FILL)
FILL_OUT = (TTY_OUT and FILL)

CURSOR_SAVE     = ctrl_seq('s')
CURSOR_RESTORE  = ctrl_seq('u')
CURSOR_HIDE     = ctrl_seq('?25l')
CURSOR_SHOW     = ctrl_seq('?25h')
CURSOR_REPORT   = ctrl_seq('6n') # '\x1B[{x};{y}R' appears as if typed into the terminal.

ALT_ENTER = ctrl_seq('?1049h')
ALT_EXIT  = ctrl_seq('?1049l')


def term_pos(x: int, y: int) -> str:
  '''
  position the cursor using 0-indexed x, y integer coordinates.
  (supposedly the 'f' suffix does the same thing).
  '''
  return ctrl_seq('H', y + 1, x + 1)


def show_cursor(show_cursor: Any) -> str:
  return CURSOR_SHOW if show_cursor else CURSOR_HIDE


def sanitize_for_console(*text:str, allow_sgr=False, allow_tab=False, escape=sgr(INVERT), unescape=sgr(RST_INVERT)) -> List[str]:
  sanitized = []
  for t in text:
    for m in _sanitize_re.finditer(t):
      s = m[0]
      k = m.lastgroup
      if k == 'vis' or (allow_sgr and k == 'sgr') or (allow_tab and k == 'tab'):
        sanitized.append(s)
      else: # Sanitize.
        sanitized.append(f'{escape}{escape_char_for_console(s)}{unescape}')
  return sanitized


_sanitize_re = _re.compile(r'''(?x)
  (?P<vis> [\n -~]+ )
| (?P<sgr> \x1b (?= \[ [\d;]* m ))
| (?P<tab> \t )
| .
''')


def escape_char_for_console(char:str) -> str:
  'Escape characters using ploy syntax.'
  return escape_reprs.get(char) or f'\\{ord(char):x};'

escape_reprs = {
  '\r': '\\r',
  '\t': '\\t',
  '\v': '\\v',
}



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from argparse import ArgumentParser
from http.server import BaseHTTPRequestHandler, HTTPServer
from sys import stdin
from typing import BinaryIO

from ..task import run


def main() -> None:

  parser = ArgumentParser(description='Serve a file or `stdin` to a browser.')
  parser.add_argument('-chrome',  action='store_true', help='Use Google Chrome')
  parser.add_argument('-firefox', action='store_true', help='Use Firefox')
  parser.add_argument('-safari',  action='store_true', help='Use Safari')
  parser.add_argument('-stp',  action='store_true', help='Use Safari Technology Preview')
  parser.add_argument('path', nargs='?', help='The file path to serve (defaults to stdin).')
  args = parser.parse_args()

  address = ('localhost', 8000) # TODO: argparse option.
  host, port = address
  addr_str = f'http://{host}:{port}'

  f_in:BinaryIO
  if args.path is not None:
    f_in = open(args.path, 'rb')
  else:
    f_in = stdin.detach() # type: ignore


  class Handler(BaseHTTPRequestHandler):

    def do_HEAD(self):
      self.send_response(200)
      self.send_header('Content-Type', 'text/html')
      self.end_headers()

    def do_GET(self):
      self.send_response(200)
      self.send_header('Content-Type', 'text/html')
      self.end_headers()
      for line in f_in:
        self.wfile.write(line)
        self.wfile.flush()


  server = HTTPServer(address, Handler)

  # note: the way we tell the OS to open the URL in the browser is a rather suspicious hack:
  # the `open` command returns and then we launch the web server,
  # relying on the fact that together the OS and the launched browser take more time to initiate the request
  # after the `open` process completes than the server does to initialize.
  if args.chrome:     run(['open', '-a', 'google chrome', addr_str])
  elif args.firefox:  run(['open', '-a', 'firefox',       addr_str])
  elif args.safari:   run(['open', '-a', 'safari',        addr_str])
  elif args.stp:      run(['open', '-a', 'safari technology preview', addr_str])
  else:               run(['open', addr_str])

  server.handle_request()


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import signal
from typing import Callable, ContextManager

from .typing import OptBaseExc, OptTraceback, OptTypeBaseExc


class Timeout(Exception):
  'Exception indicating that an Alarm timed out.'


class Alarm(ContextManager):

  def __init__(self, timeout: int, msg:str='alarm timed out after {timeout} seconds', on_signal: Callable[[], None]=None) -> None:
    if not isinstance(timeout, int): raise TypeError(f'timeout must be an int; received: {timeout!r}')
    if timeout < 0: raise TypeError(f'timeout must be nonnegative; received: {timeout!r}')
    self.timeout = timeout
    self.msg = msg
    self.on_signal = on_signal
    self.timed_out = False


  def __enter__(self) -> None:
    if not self.timeout: return
    prev = signal.getsignal(signal.SIGALRM)
    if prev is not signal.SIG_DFL:
      raise Exception(f'task.run encountered previously installed signal handler: {prev}')

    def alarm_handler(signum, current_stack_frame) -> None:
      # since signal handlers carry reentrancy concerns; do not do any IO within the handler.
      self.timed_out = True
      if self.on_signal: self.on_signal() # callback must respect reentrancy limitations.

    signal.signal(signal.SIGALRM, alarm_handler) # set handler.
    signal.alarm(self.timeout) # set alarm.


  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    if not self.timeout: return
    signal.alarm(0) # disable alarm.
    signal.signal(signal.SIGALRM, signal.SIG_DFL) # uninstall handler.
    if exc_type: return # exception will be reraised.
    if self.timed_out: raise Timeout(self.msg.format(timeout=self.timeout))



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from functools import total_ordering
from io import TextIOBase, TextIOWrapper
from tarfile import TarFile
from typing import Any, BinaryIO, Callable, Dict, Iterable, List, Type, Union, cast
from zipfile import BadZipFile, ZipFile

from .fs import path_ext
from .util import lazy_property


FileOrPath = Union[BinaryIO, str] # TODO: support other pathlike objects.
Opener = Callable[[], BinaryIO]


class Archive:

  def __init__(self, file_or_path:FileOrPath, ext=None) -> None:
    self.name: str = getattr(file_or_path, 'name', str(file_or_path))
    if ext is None:
      ext = path_ext(self.name)
    self.ext = ext
    try: self._handler = _ext_handlers[ext](file_or_path)
    except KeyError: raise ValueError(f'Archive does not support extension: {ext!r}')

  def __repr__(self) -> str:
    return f'{self.__class__.__name__}(name={self.name!r})'

  def __iter__(self) -> Iterable['ArchiveFile']:
    return self._handler.files(self)

  @property
  def file_names(self) -> List[str]: return self._handler.file_names



@total_ordering
class ArchiveMember:

  def __init__(self, archive: Archive, name: str) -> None:
    self.archive = archive
    self.name = name

  def __repr__(self) -> str:
    return f'{self.__class__.__name__}({self.archive}, {self.name!r})'

  def __lt__(self, other: Any) -> Union[bool, 'NotImplemented']:
    if isinstance(other, ArchiveMember):
      return self.name < other.name
    return NotImplemented



class ArchiveFile(ArchiveMember):

  def __init__(self, archive:Archive, name:str, opener:Opener) -> None:
    super().__init__(archive=archive, name=name)
    self._opener = opener

  @lazy_property
  def _file(self): return self._opener()

  def __iter__(self): return iter(self._file)

  def __getattr__(self, name):
    # note: dunder attribute access is never forwarded to this method.
    if name.startswith('_'): raise AttributeError(name)
    return getattr(self._file, name)

  def seekable(self):
    return getattr(self._file, 'seekable', False)

  def text(self, encoding='UTF-8', errors=None, newline=None) -> TextIOWrapper:
    'Returns a TextIOWrapper around the underlying binary buffer.'
    return TextIOWrapper(self._file, encoding=encoding, errors=errors, newline=newline)



class ArchiveDir(ArchiveMember): pass



class _Handler:

  def __init__(self, file_or_path:FileOrPath) -> None: raise NotImplementedError

  def files(self, archive:Archive) -> Iterable[ArchiveFile]: raise NotImplementedError

  @property
  def file_names(self) -> List[str]: raise NotImplementedError


class _TarHandler(_Handler):
  def __init__(self, file_or_path:FileOrPath) -> None:
    if isinstance(file_or_path, str):
      file_or_path = open(file_or_path, 'rb')
    self.tar = TarFile.open(mode='r|*', fileobj=file_or_path)

  def files(self, archive:Archive) -> Iterable[ArchiveFile]:
    for info in self.tar:
      if info.isdir(): continue
      def opener() -> BinaryIO:
        f = self.tar.extractfile(info)
        if f is None: raise Exception(f'archive failed to open file: {info.name}')
        return cast(BinaryIO, f)
      #^ note: crucial to pass info here and not name, or else gz archives get very slow.
      yield ArchiveFile(archive=archive, name=info.name, opener=opener)

  @property
  def file_names(self) -> List[str]: return self.tar.getnames()



class _ZipHandler(_Handler):
  def __init__(self, file_or_path:FileOrPath) -> None:
    try: self.zip = ZipFile(file_or_path)
    except BadZipFile as e:
      if isinstance(file_or_path, TextIOBase): # type: ignore
        raise TypeError(f'Archive requires a path or binary stream but received text stream: {file_or_path}') from e
      raise

  def files(self, archive:Archive) -> Iterable[ArchiveFile]:
    for name in self.file_names:
      if name.endswith('/'): continue # directory.
      def opener() -> BinaryIO: return cast(BinaryIO, self.zip.open(name, 'r')) # mypy bug: BinaryIO and IO[bytes] should be equivalent.
      yield ArchiveFile(archive=archive, name=name, opener=opener)

  @property
  def file_names(self) -> List[str]: return self.zip.namelist()



_ext_handlers: Dict[str, Type[_Handler]] = {
  '.bz2' : _TarHandler,
  '.gz'  : _TarHandler,
  '.tar' : _TarHandler,
  '.xz'  : _TarHandler,
  '.zip' : _ZipHandler,
}



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from dataclasses import dataclass
from typing import Any, Iterable, Iterator, Tuple


SyntaxStack = Tuple['Syntax', ...]


@dataclass(frozen=True)
class Syntax:

  @dataclass(frozen=True)
  class Pos:
    '''
    All indices are zero-based.
    '''
    line: int
    col: int
    end_line: int
    end_col: int

    def __init__(self, *, line=-1, col=-1, end_line=-1, end_col=-1, enclosed:Iterable[Any]=()) -> None:
      for el in enclosed:
        if el is None: continue
        pos = el if isinstance(el, Syntax.Pos) else el.pos # `enclosed` elements should either be a `Pos` or have a `pos`.
        if pos.line >= 0:
          if line < 0 or line > pos.line: line = pos.line
          if pos.col >= 0 and line == pos.line and (col < 0 or col > pos.col): col = pos.col
          if pos.end_line >= 0:
            if end_line < pos.end_line: end_line = pos.end_line
            if pos.end_col >= 0 and end_line == pos.end_line and end_col < pos.end_col: end_col = pos.end_col
      set_ = super().__setattr__
      set_('line', line)
      set_('col', col)
      set_('end_line', end_line)
      set_('end_col', end_col)

    def __str__(self): return f'{self.line+1:04}:{self.col+1:03}-{self.end_line+1:04}:{self.end_col+1:03}'

    def expand(self, pos:'Syntax.Pos') -> 'Syntax.Pos':
      line = self.line
      col = self.col
      end_line = self.end_line
      end_col = self.end_col
      if line < 0 or line > pos.line: line = pos.line
      if line == pos.line and (col < 0 or col > pos.col): col = pos.col
      if end_line < pos.end_line: end_line = pos.end_line
      if end_line == pos.end_line and end_col < pos.end_col: end_col = pos.end_col
      return Syntax.Pos(line=line, col=col, end_line=end_line, end_col=end_col)

  @staticmethod
  def path_with_slash(path:str) -> str:
    '''
    'VSCode does not recognize paths unless they contain a slash,
    so add a leading `./` to any simple name that does not look special.
    '''
    return path if ('<' in path or '/' in path) else './' + path

  @staticmethod
  def diagnostic(path:str, *, line:int, col:int, end_line=-1, end_col=-1, msg='') -> str:
    s = ' ' if msg else ''
    return f'{Syntax.path_with_slash(path)}:{Syntax.Pos(line=line, col=col, end_line=end_line, end_col=end_col)}:{s}{msg}'

  path: str
  pos: Pos
  kind: str
  content: Any


  @property
  def children(self) -> Iterator['Syntax']:
    if isinstance(self.content, str): return iter(())
    try: return iter(self.content)
    except TypeError: return iter(())

  def get_children(self) -> Iterator['Syntax']:
    'Function alias of the `children` property. Provided for ease of use with `pithy.tree.transform_tree`.'
    return self.children


  def describe(self, *, prefix='', depth=0) -> Iterator[str]:
    'Generate lines (without terminating newlines) describing the receiver.'
    spaces = '  ' * depth
    kind = f' {self.kind}:' if self.kind else ''
    content = self.content
    if isinstance(content, str):
      if len(content) <= 64:
        text = f' {content!r}'
      else:
        text = f' {content[:32]!r} … {content[-32:]!r}'
    else:
      text = ''
    yield f'{prefix}{Syntax.path_with_slash(self.path)}:{self.pos}:{spaces}{kind}{text}'
    for child in self.children:
      yield from child.describe(prefix=prefix, depth=depth+1)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from sys import stdout
from typing import Any, BinaryIO, Callable, Dict, List, Tuple

from msgpack import ExtraData, FormatError, OutOfData, StackError, Unpacker, dump as _dump, load as _load  # type: ignore

from .encode import EncodeObj, encode_obj

ObjectHook = Callable[[Dict[str,Any]], Any]
ObjectPairsHook = Callable[[List[Tuple[str,Any]]], Any]
ListHook = Callable[[List[Any]], Any]
ExtHook = Any # TODO

_ml = 2147483647


def load_msgpack(file:BinaryIO, use_list=False, raw=False, strict_map_key=False,
 object_hook:ObjectHook=None, object_pairs_hook:ObjectPairsHook=None, list_hook:ListHook=None,
 encoding:str=None, unicode_errors:str=None, ext_hook:ExtHook=None,
 max_str_len=_ml, max_bin_len=_ml, max_array_len=_ml, max_map_len=_ml, max_ext_len=_ml) -> Any:
  # Omitted: read_size=0, max_buffer_size=0.
  return _load(file, use_list=use_list, raw=raw, strict_map_key=strict_map_key,
    object_hook=object_hook, object_pairs_hook=object_pairs_hook, list_hook=list_hook,
    encoding=encoding, unicode_errors=unicode_errors, ext_hook=ext_hook,
    max_str_len=max_str_len, max_bin_len=max_bin_len, max_array_len=max_array_len,
    max_map_len=max_map_len, max_ext_len=max_ext_len)


def load_msgpacks(file:BinaryIO, use_list=False, raw=False, strict_map_key=False,
 object_hook:ObjectHook=None, object_pairs_hook:ObjectPairsHook=None, list_hook:ListHook=None,
 encoding:str=None, unicode_errors:str=None, ext_hook=None,
 max_str_len=_ml, max_bin_len=_ml, max_array_len=_ml, max_map_len=_ml, max_ext_len=_ml) -> Any:

  return Unpacker(file, use_list=use_list, raw=raw, strict_map_key=strict_map_key,
    object_hook=object_hook, object_pairs_hook=object_pairs_hook, list_hook=list_hook,
    encoding=encoding, unicode_errors=unicode_errors, ext_hook=ext_hook,
    max_str_len=max_str_len, max_bin_len=max_bin_len, max_array_len=max_array_len,
    max_map_len=max_map_len, max_ext_len=max_ext_len)


def count_msgpacks(file:BinaryIO) -> int:
  unpacker = Unpacker(file)
  i = 0
  while True:
    try: unpacker.skip()
    except OutOfData: return i
    i += 1


def write_msgpack(file:BinaryIO, obj:Any, default:EncodeObj=encode_obj, unicode_errors='strict',
 use_single_float=False, autoreset=False, use_bin_type=True, strict_types=False) -> None:

  _dump(obj, file, default=encode_obj, unicode_errors=unicode_errors,
    use_single_float=use_single_float, use_bin_type=use_bin_type, strict_types=strict_types)


def out_msgpack(obj:Any, default:EncodeObj=encode_obj, unicode_errors='strict',
 use_single_float=False, autoreset=False, use_bin_type=True, strict_types=False) -> None:

  _dump(obj, stdout.buffer, default=encode_obj, unicode_errors=unicode_errors,
    use_single_float=use_single_float, use_bin_type=use_bin_type, strict_types=strict_types)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
XML tools.
'''

from io import StringIO
from itertools import chain, count
from typing import Any, ContextManager, Dict, Optional, TextIO, Tuple, Type, TypeVar

from ..typing import OptBaseExc, OptTraceback, OptTypeBaseExc

from .escape import EscapedStr, XmlAttrs, esc_xml_text, fmt_attrs


_XmlWriter = TypeVar('_XmlWriter', bound='XmlWriter')


class IndexCounters:
  'Internal pair of counters for XmlWriter to generate "id" and "class" attributes.'
  def __init__(self) -> None:
    self.id_counter = count()
    self.class_counter = count()


class XmlWriter(ContextManager):
  '''
  XmlWriter is the root class for building document trees for output.
  XmlWriter is subclassed to provide more convenient Python APIs; see pithy.html and pithy.svg.
  '''

  replaced_attrs = {
    'cl' : 'class',
  }

  can_auto_close_tags = True # Allows treating all elements as "void" or "self-closing", e.g. '<TAG/>'. False for HTML.
  is_self_closing = False # True for e.g. HTML tags that are self closing, such as <img />.
  tag:str = '' # Subclasses can specify a tag.

  def __init__(self, *_children:Any, tag:str=None, _counters:IndexCounters=None, attrs:XmlAttrs=None, **kwargs:Any) -> None:
    '''
    `attrs` also allows for XML attributes that contain non-identifier characters.
    '''
    self.tag = tag or type(self).tag
    if not self.tag: raise Exception(f'{type(self)}: neither type-level tag nor constructor tag specified')
    self.prefix = ''
    self.children = list(_children)
    self.inline:Optional[bool] = None
    self.attrs = kwargs # Guaranteed not to be referenced anywhere else, whereas attrs might be.
    if attrs: self.attrs.update(attrs)
    self._appears_inline = bool(self.children)
    self._context_depth = 0
    self._is_closed = False
    self._counters:IndexCounters = _counters or IndexCounters()


  def write(self, file:TextIO, end='\n') -> None:
    if self.prefix: print(self.prefix, file=file)
    print('<', self.tag, fmt_attrs(self.attrs, self.replaced_attrs), sep='', end='', file=file)
    if self.can_auto_close_tags and not self.children:
      print('/>', end=end, file=file)
      return
    inline = self._appears_inline if (self.inline is None) else self.inline
    child_end = ('' if inline else '\n')
    print('>', end=child_end, file=file)
    for child in self.children:
      if isinstance(child, XmlWriter):
        child.write(end=child_end, file=file)
      else:
        print(esc_xml_text(child), end=child_end, file=file)
    print('</', self.tag, '>', sep='', end=end, file=file)


  @property
  def string(self) -> str:
    s = StringIO()
    self.write(file=s)
    return s.getvalue()


  def __repr__(self) -> str:
    return f'{self.__class__.__name__}({self.tag!r})'


  def __enter__(self:_XmlWriter) -> _XmlWriter:
    if self._is_closed: raise Exception(f'XmlWriter is already closed: {self}')
    self._context_depth += 1
    return self

  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    self._context_depth -= 1
    if self._context_depth == 0:
      self._is_closed = True
      self.on_close()


  def on_close(self) -> None:
    'Optional closing action for subclasses.'


  def add(self, *items:Any, sep='', end='\n') -> None:
    if self._is_closed: raise Exception(f'XmlWriter is already closed: {self}')
    self.children.extend(items)


  def leaf(self, tag:str, *, attrs:XmlAttrs) -> None:
    self.add(EscapedStr(f'<{tag}{fmt_attrs(attrs, self.replaced_attrs)}/>'))


  def child(self, child_class:Type[_XmlWriter], *children:Any, attrs:XmlAttrs=None, **kwargs:Any) -> _XmlWriter:
    'Create a child XmlWriter for use in a `with` context to represent a nesting XML element.'
    if self._is_closed: raise Exception(f'XmlWriter is already closed: {self}')
    c = child_class(*children, _counters=self._counters, attrs=attrs, **kwargs)
    self.add(c)
    return c


  def gen_id(self) -> str:
    return f'_id{next(self._counters.id_counter)}'

  def gen_class(self) -> str:
    return f'_class{next(self._counters.class_counter)}'


def add_opt_attrs(attrs:Dict[str,Any], *pairs:Tuple[str, Any], **items:Any) -> None:
  'Add the items in `*pairs` and `**items` attrs, excluding any None values.'
  for k, v in chain(pairs, items.items()):
    if v is None: continue
    assert k not in attrs, k
    attrs[k] = v



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Optional, Type
from types import TracebackType

Opt = Optional

# These types are helpful when defining `__exit__`:
# def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> bool: ...
OptTypeBaseExc = Optional[Type[BaseException]]
OptBaseExc = Optional[BaseException]
OptTraceback = Optional[TracebackType]



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from urllib.parse import urlsplit as split_url, urlunsplit as compose_url

from .path import path_stem


def url_scheme(url:str) -> str: return split_url(url).scheme

def url_netloc(url:str) -> str: return split_url(url).netloc

def url_path(url:str) -> str: return split_url(url).path

def url_query(url:str) -> str: return split_url(url).query

def url_fragment(url:str) -> str: return split_url(url).fragment


def url_scheme_netloc_path(url:str) -> str:
  scheme, netloc, path, query, fragment = split_url(url)
  return compose_url((scheme, netloc, path, None, None))


def url_netloc_path(url:str) -> str:
  scheme, netloc, path, query, fragment = split_url(url)
  return compose_url((None, netloc, path, None, None))


def drop_url_fragment(url:str) -> str:
  scheme, netloc, path, query, fragment = split_url(url)
  return compose_url((scheme, netloc, path, query, None))


def replace_url_ext(url:str, ext:str) -> str:
  scheme, netloc, path, query, fragment = split_url(url)
  stem = path_stem(path)
  return compose_url((scheme, netloc, stem + ext, query, fragment))




# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'Sequence utilities.'


from bisect import bisect_left, bisect_right
from typing import Iterator, Sequence, Tuple, TypeVar


_T = TypeVar('_T')


def sorted_seq_index(seq:Sequence[_T], el:_T) -> int:
  '''
  Locate the leftmost element exactly equal to `el`.
  From stdlib bisect documentation.
  '''
  i = bisect_left(seq, el)
  if i != len(seq) and seq[i] == el:
    return i
  raise ValueError(el)


def sorted_seq_find_lt(seq:Sequence[_T], el:_T) -> _T:
  '''
  Find rightmost element less than `el`.
  From stdlib bisect documentation.
  '''
  i = bisect_left(seq, el)
  if i:
    return seq[i-1]
  raise ValueError


def sorted_seq_find_le(seq:Sequence[_T], el:_T) -> _T:
  '''
  Find rightmost element less than or equal to `el`.
  From stdlib bisect documentation.
  '''
  i = bisect_right(seq, el)
  if i:
    return seq[i-1]
  raise ValueError


def sorted_seq_find_gt(seq:Sequence[_T], el:_T) -> _T:
  '''
  Find leftmost element greater than `el`.
  From stdlib bisect documentation.
  '''
  i = bisect_right(seq, el)
  if i != len(seq):
    return seq[i]
  raise ValueError


def sorted_seq_find_ge(seq:Sequence[_T], el:_T) -> _T:
  '''Find leftmost element greater than or equal to `el`.
  From stdlib bisect documentation.
  '''
  i = bisect_left(seq, el)
  if i != len(seq):
    return seq[i]
  raise ValueError


def iter_pairs_of_el_is_last(seq:Sequence[_T]) -> Iterator[Tuple[_T, bool]]:
  last = len(seq) - 1
  return ((el, i==last) for i, el in enumerate(seq))



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import inspect
from inspect import FrameInfo
from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, TypeVar, cast


def bindings_matching(*, prefix:str=None, val_type:type=None, strip_prefix=True, frame='<module>') -> List[Tuple[str, Any]]:
  '''
  Return (name, value) pairs of bindings from the specified frame,
  that match the specified prefix and val_type filters.
  Frame must be either an int (depth; immediate caller is 1),
  or a string (the name of the target frame's function, or '<module>', the default).
  '''
  stack = inspect.stack()
  info: Optional[FrameInfo]
  if isinstance(frame, int):
    info = stack[frame]
  elif isinstance(frame, str):
    info = None
    for info in stack:
      if info.function == frame:
        break
    if info is None:
      raise ValueError('call stack does not contain a matching frame: {}'.format(frame))
  else:
    raise TypeError("frame parameter must be either an int (depth; immediate caller is 1), "
      "or a string (the name of the target frame's function, or '<module>', the default).")
  bindings = info.frame.f_locals
  pairs = []
  for name, value in bindings.items():
    if all((
      prefix is None or name.startswith(prefix),
      val_type is None or isinstance(value, val_type),
    )):
      if prefix and strip_prefix:
        name = name[len(prefix):]
      pairs.append((name, value))
  return pairs


def dispatcher_for_defs(*, prefix:str, default:Callable=None, base:Mapping[str,Callable]={}, exclude:Iterable[str]=()) -> Callable:
  '''
  Creates a dispatcher function from callable definitions in the caller's module.
  Only callables whose name starts with `prefix` are included.
  `default` is another callable; it will receive all of the arguments to the dispatcher (including the first string argument)
  when no particular dispatch function matches.
  '''

  _exclude = {exclude} if isinstance(exclude, str) else set(exclude)

  bindings:Dict[str,Callable] = dict(base)
  for name, fn in bindings_matching(prefix=prefix, frame='<module>', strip_prefix=True):
    if (name in bindings) or (name in _exclude) or (not callable(fn)): continue
    bindings[name] = fn

  if default is None:
    def dispatch_fn(name, *args, **kwargs):
      fn = bindings[name]
      try: return fn(*args, **kwargs)
      except Exception as exc: raise DispatchWrapperException(fn) from exc
  else:
    def dispatch_fn(name, *args, **kwargs):
      try: fn = bindings[name]
      except KeyError: pass # Handled below.
      else:
        try: return fn(*args, **kwargs)
        except Exception as exc: raise DispatchWrapperException(fn) from exc
      return default(name, *args,  **kwargs) # type: ignore

  setattr(dispatch_fn, 'bindings', bindings)
  setattr(dispatch_fn, 'default', default)
  return dispatch_fn


class DispatchWrapperException(Exception): pass


_A = TypeVar('_A', bound=Any)

def rename(obj:_A, name:str=None, module:str=None) -> _A:
  'Returns `obj`, after renaming name and/or module.'
  if name is not None:
    obj.__name__ = name
    obj.__qualname__ = name
  if module is not None:
    obj.__module__ = module
  return obj


def get_caller_module_name(steps=1) -> Optional[str]:
  'Get the module name of the caller name, `steps` number of frames from the immediate caller (defaults to 1).'
  f = inspect.currentframe() # This frame.
  if f is None: return None
  f = f.f_back # Immediate caller's frame; caller already knows this.
  if f is None: return None
  for i in range(steps):
    f = f.f_back
    if f is None: return None
  spec = f.f_globals['__spec__']
  if spec: return cast(str, spec.name)
  else: return cast(str, f.f_globals['__name__'])


def main_file_path() -> str:
  'Returns the main file path.'
  import __main__ # type: ignore # mypy bug.
  path: str = __main__.__file__
  return path



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from itertools import chain
from typing import Dict, Iterable, Tuple


# use pairs instead of real range objects because they are sortable, and faster to load in the interpreter.
CodeRange = Tuple[int, int]
CodeRanges = Tuple[CodeRange, ...]


# These (real) ranges are provided merely as a reference / convenience to the user.
ascii_range = range(0x80)
unicode_range = range(0x110000)

high_surrogates = (0xD800, 0xDC00)
low_surrogates  = (0xDC00, 0xE000)
surrogates = (high_surrogates[0], low_surrogates[1])


planes:Tuple[CodeRanges, ...] = (
  ( # 0: Basic Multilingual Plane.
    # Note: the surrogates range is excluded because those code points are not legally encodable.
    (0x0000, surrogates[0]),
    (surrogates[1], 0x10000),
  ),
  ( # 1: Supplementary Multilingual Plane.
    (0x10000, 0x15000),
    (0x16000, 0x19000),
    (0x1B000, 0x1C000),
    (0x1D000, 0x20000),
  ),
  ( # 2: Supplementary Ideographic Plane.
    (0x20000, 0x2D000),
    (0x2F000, 0x30000),
  ),
  (), # 3: Unassigned.
  (), # 4: Unassigned.
  (), # 5: Unassigned.
  (), # 6: Unassigned.
  (), # 7: Unassigned.
  (), # 8: Unassigned.
  (), # 9: Unassigned.
  (), # 10: Unassigned.
  (), # 11: Unassigned.
  (), # 12: Unassigned.
  (), # 13: Unassigned.
  ( # 14: Supplement­ary Special-purpose Plane.
    (0xE0000, 0xE1000),
  ),
  ( # 15: Supplement­ary Private Use Area Plane A (SPUA-A).
    (0xF0000, 0x100000),
  ),
  ( # 16: Supplement­ary Private Use Area Plane A (SPUA-B).
    (0x100000, 0x110000),
  ),
)

abbreviated_planes:Dict[str, CodeRanges]  = {
  'BMP': planes[0],
  'SMP': planes[1],
  'SIP': planes[2],
  'SSP': planes[14],
  'SPUA_A': planes[15],
  'SPUA_B': planes[16],
}

all_plane_ranges = tuple(chain(*(planes)))


def codes_for_ranges(seq:Iterable[CodeRange]) -> Iterable[int]:
  return chain.from_iterable(range(*r) for r in seq)


def ranges_for_codes(seq:Iterable[int]) -> Iterable[CodeRange]:
  'codes must be sorted.'
  it = iter(seq)
  try: first = next(it)
  except StopIteration: return
  low = first
  end = first + 1
  for el in it:
    if el < end: raise ValueError(el)
    if el == end:
      end += 1
    else:
      yield (low, end)
      low = el
      end = el + 1
  yield (low, end)


def coalesce_sorted_ranges(seq:Iterable[CodeRange]) -> Iterable[CodeRange]:
  it = iter(seq)
  try: low, end = next(it)
  except StopIteration: return
  for r in it:
    l, e = r
    if e < l: raise ValueError(r) # bad range element.
    if l < low:
      raise ValueError(f'coalesce_sorted_ranges encountered unsorted range element: {r!r}; current coalesced range: {(low, end)}')
    if l <= end:
      end = max(end, e)
    else:
      yield (low, end)
      low = l
      end = e
  yield (low, end)


def union_sorted_ranges(*seqs:Iterable[CodeRange]) -> Iterable[CodeRange]:
  return coalesce_sorted_ranges(sorted(chain(*seqs)))


def intersect_sorted_ranges(seq_a:Iterable[CodeRange], seq_b:Iterable[CodeRange]) -> Iterable[CodeRange]:
  iter_a = iter(seq_a)
  iter_b = iter(seq_b)
  try: # iteration scope.
    a, ae = next(iter_a)
    b, be = next(iter_b)
    while True:
      while ae <= b: # drop a.
        a, ae = next(iter_a)
      if be <= a: # drop b.
        b, be = next(iter_b)
        continue # must continue from top, to retest a.
      s = max(a, b)
      if ae <= be:
        yield (s, ae)
        b, be = (ae, be) # if b is empty it will get dropped on next pass, assuming seq_a is coalesced.
      else:
        yield (s, be)
        a, ae = (be, ae) # if b is empty it will get dropped on next pass, assuming seq_b is coalesced.
  except StopIteration: return




# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Generic dictionary trees for XML.

Trees are represented as ordinary dictionaries:
* The element tag is stored under the '' key;
* attributes are stored under string keys;
* node children and text are stored together under zero-indexed numeric keys.

Note that this scheme relies on the stable ordering of dict items, so it requires Python3.6+.
'''

import re
from typing import Any, Callable, Container, Dict, Iterator, List, Tuple, Union, cast
from xml.etree.ElementTree import Element

from lxml.etree import Comment  # type: ignore

from ..exceptions import DeleteNode
from .escape import fmt_attr_items


GenericXmlKey = Union[int,str]
GenericXmlVal = Union[str,Dict]
GenericXml = Dict[GenericXmlKey,GenericXmlVal]


def generic_xml_from_etree(el:Element) -> GenericXml:
  tag = el.tag
  if tag == Comment: tag = '!COMMENT'
  res:Dict = {'': tag}
  res.update(sorted(el.items()))
  idx = 0
  t = el.text
  if t:
    res[idx] = t
    idx += 1
  for child in el:
    res[idx] = generic_xml_from_etree(child)
    idx += 1
    t = child.tail
    if t:
      res[idx] = t
      idx += 1
  return res


def get_tag_attrs_children(xml:GenericXml) -> Tuple[str, List[Tuple[str,Any]], List[Any]]:
  tag = cast(str, xml[''])
  attrs:List[Tuple[str,Any]] = []
  children:List[Tuple[int,Any]] = []
  for p in xml.items():
    k = p[0]
    if isinstance(k, str):
      if k: attrs.append(p) # type: ignore
    else:
      children.append(p) # type: ignore
  return tag, attrs, children


def get_text(xml:GenericXml) -> str:
  text:List[str] = []
  for k, v in xml.items():
    if isinstance(k, int):
      if isinstance(v, dict): text.append(get_text(v))
      else: text.append(str(v))
  return ''.join(text)


def clean_generic_xml_whitespace(xml:GenericXml, ws_sensitive_tags:Container[str]) -> None:

  def _clean(xml:GenericXml) -> None:
    tag = xml['']
    children:List[GenericXmlVal] = []

    # Get all children, consolidating consecutive strings, and simultaneously remove all children from the node.
    for k, v in tuple(xml.items()):
      if isinstance(k, str): continue # Leave tag and attrs untouched.
      del xml[k]
      if isinstance(v, str): # Text.
        if not v: continue # Omit empty strings.
        if children and isinstance(children[-1], str): # Consolidate.
          children[-1] += v
        else:
          children.append(v)
      else: # Child element.
        children.append(v)

    if tag not in ws_sensitive_tags: # Clean whitespace.
      for i in range(len(children)):
        v = children[i]
        if not isinstance(v, str): continue
        replacement = '\n' if '\n' in v else ' '
        children[i] = ws_re.sub(replacement, v)

    xml.update(enumerate(children)) # Replace children with fresh, compacted indices.

  visit_generic_xml(xml, _clean)

ws_re = re.compile(r'\s+')


def render_generic_xml(xml:GenericXml, void_elements:Container[str]) -> Iterator[str]:
  tag, attrs, children = get_tag_attrs_children(xml)

  if (not children) and void_elements:
    self_closing = (tag in void_elements)
  else:
    self_closing = (not children)

  attrs_str = fmt_attr_items(attrs, {})
  head = f'<{tag}{attrs_str}'

  if self_closing:
    yield head + '/>'
  else:
    yield head + '>'
    for _, child in children:
      if isinstance(child, dict):
        yield from render_generic_xml(child, void_elements=void_elements)
      else:
        yield str(child)
    yield f'</{tag}>'


def visit_generic_xml(xml:GenericXml, visit:Callable[[GenericXml],None]) -> None:
  visit(xml)
  for k, v in tuple(xml.items()):
    if isinstance(k, int) and isinstance(v, dict): # Child element.
      try: visit_generic_xml(v, visit)
      except DeleteNode: del xml[k]




# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

# This is the setup.py for for all packages in the repo.
# The package to be installed is specified by setting the PITHY_PACKAGE environment variable,
# or else inferred from the name of the parent directory (during end-user pip install).
# Each package has its own `setup.cfg`.

from configparser import ConfigParser
from distutils.command.build_scripts import build_scripts  # type: ignore
from itertools import chain
from os import chmod, environ, getcwd as current_dir, listdir as list_dir, mkdir as make_dir, walk
from os.path import (basename as path_name, dirname as path_dir, exists as path_exists, isdir as is_dir, join as path_join,
  normpath as norm_path, split as split_dir_name, splitext as split_ext)
from pprint import pprint
from typing import Any, List

from setuptools import Command, setup  # type: ignore
from setuptools.command.develop import develop  # type: ignore
from setuptools.command.install import install  # type: ignore
from setuptools.command.install_scripts import install_scripts  # type: ignore
from setuptools.config import read_configuration  # type: ignore


# We stubbornly insist on keeping each package directory at the root of the repo.
# Since we want to distribute each package separately, we need a setup.py for each.
# Pip requires that the setup script is named setup.py,
# and dev mode (`-e`) requires that we place setup.py in the parent directory of each package.
# Therefore, we must the same setup.py for each package.
# We make this work by telling setup.py which package it is installing via the PITHY_PACKAGE environment variable.

base_dir = path_dir(norm_path(__file__)) or '.' # When run from pip, this is a dir with the package name, inside a temp dir.
base_name = path_name(base_dir)
assert not base_dir.endswith('/'), base_dir
base_dir += '/'
base_prefix_len = len(base_dir)

package_name = environ.get('PITHY_PACKAGE')
if package_name is None:
  print(f'PITHY_PACKAGE environment variable not set; using: {base_name!r}')
assert package_name
assert not package_name.endswith('/'), package_name

package_dir = path_join(base_dir, package_name)
bin_src_dir = path_join(package_dir, 'bin')


def msg(*items:Any) -> None: print(' ', *items)


class BuildScripts(build_scripts): # type: ignore
  def run(self) -> None:
    super().run()


class Develop(develop): # type: ignore
  def run(self) -> None:
    super().run()
    install_bins(dst_dir=self.script_dir)


class Install(install): # type: ignore
  def run(self) -> None:
    super().run()


class InstallScripts(install_scripts): # type: ignore
  def run(self) -> None:
    msg('InstallScripts')
    super().run()
    install_bins(dst_dir=self.install_dir)


def install_bins(dst_dir:str) -> None:
  '''
  Generate executable script entry points.
  We do this because standard entry_points/console_scripts have noticeably slow startup times,
  apparently due to overly complex boilerplate.
  '''
  msg('bin_src_dir:', bin_src_dir)
  msg('bin dst_dir:', dst_dir)

  if not is_dir(bin_src_dir):
    msg('note: no bin directory.')
    return
  if not path_exists(dst_dir): make_dir(dst_dir)
  py_path = path_join(dst_dir, 'python3')
  for name in list_dir(bin_src_dir):
    stem, ext = split_ext(name)
    if ext != '.py' or stem.startswith('.') or stem.startswith('_'): continue
    path = path_join(dst_dir, stem.replace('_', '-')) # Omit extension from bin name and use dashes.
    module = f'{package_name}.bin.{stem}'
    msg(f'generating script: {path}')
    with open(path, 'w') as f:
      f.write(bin_template.format(py_path=py_path, module=module))
      chmod(f.fileno(), 0o755)

bin_template = '''\
#!{py_path}
# Generated by pithy/setup.py.
from {module} import main
main()
'''


def discover_packages() -> List[str]:
  '''
  Discover subpackages by traversing over the directory tree.
  Verify that all subdirectories have `__init__` files.
  '''
  bad_names = []
  missing_inits = []
  packages = []
  for dir_path, dir_names, file_names in walk(package_dir):
    rel_path = dir_path[base_prefix_len:]
    subpackage_name = rel_path.strip('./').replace('/', '.')
    packages.append(subpackage_name)
    msg('discovered subpackage:', rel_path, '->', subpackage_name)
    # Filter the subdirectories in place, as permitted by `walk`.
    dir_names[:] = filter(is_subpackage, dir_names)
    # Validate names. Collect them so that we can issue all error messages at once, then exit.
    for name in chain(dir_names, file_names):
      if '-' in name: bad_names.append(path_join(dir_path, name))
    if '__init__.py' not in file_names:
      missing_inits.append(path_join(dir_path, '__init__.py'))

  if bad_names: msg(f'bad module names:\n' + '\n'.join(sorted(bad_names)))
  if missing_inits:
    msg(f'missing package __init__.py files:\n    ' + '\n    '.join(repr(s) for s in sorted(missing_inits)))
  if bad_names or missing_inits: exit(1)
  msg('packages:', *packages)
  return packages


def is_subpackage(dir_name:str) -> bool:
  if '.' in dir_name: return False
  if dir_name in ('__pycache__',): return False
  return True


def main() -> None:

  msg('base_dir:', base_dir)
  msg('package_dir:', package_dir)
  msg('package_name:', package_name)
  packages = discover_packages()
  py_module = package_dir + '.py'
  py_modules = [py_module] if path_exists(py_module) else []

  # The root `setup.cfg` contains only common config data for all packages; it is loaded automatically by setuptools.
  # Each package has its own `setup.cfg` which must be loaded manually.
  cfg_path = f'setup-{package_name}.cfg'
  metadata = read_configuration(cfg_path)['metadata'] # Get the metadata the setuptools way.

  if 'name' not in metadata: exit('setup.cfg: error: missing `name`.')

  args = dict(
    license='CC0',
    author='George King',
    author_email='george.w.king@gmail.com',
    url='https://github.com/gwk/pithy',
    include_package_data=True,
    packages=packages,
    cmdclass={
      'build_scripts': BuildScripts,
      'develop': Develop,
      'install': Install,
      'install_scripts': InstallScripts,
    },
    py_modules=[],
    package_dir={package_name:package_dir},
    **metadata,
  )

  #pprint(args)
  setup(**args)


if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from sys import argv, stderr, stdin


def main() -> None:
  args = argv[1:] or list(stdin)
  number_strings = [a.replace(',', '_') for arg in args for a in arg.split()]
  print(number_strings, file=stderr)
  try: total = sum(float(s) for s in number_strings)
  except ValueError as e: exit(f'sum error: {e}')
  print(total)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'Graphviz dot utliities.'

import re
from html import escape as html_escape
from sys import stdout
from typing import Any, Callable, Dict, Iterable, Optional, TextIO, Tuple, Union, Mapping


GraphvizName = Union[int, float, str]

GraphvizAttrs = Mapping[str,GraphvizName]
GraphvizValAttrs = Tuple[GraphvizName,GraphvizAttrs]
GraphvizValues = Union[Iterable[GraphvizName], Iterable[GraphvizValAttrs], Mapping[GraphvizName,GraphvizAttrs]]
GraphvizAdjacency = Union[Mapping[GraphvizName,GraphvizValues], Iterable[Tuple[GraphvizName,GraphvizValues]]]
GraphvizNodes = Mapping[GraphvizName,GraphvizAttrs]

dot_bare_id_re = re.compile(r'[_a-zA-Z][_a-zA-Z0-9]*')
dot_quotable_id_re = re.compile(r'([ !#-\[\]-~\w]| )+') # Exclude double-quote and backslash; allow non-ascii word chars.
dot_keywords = frozenset({'node', 'edge', 'graph', 'digraph', 'subgraph', 'strict'})


def dot_id_quote(name:GraphvizName) -> str:
  '''
  Properly quote an identifier.
  The DOT language has two quotation mechanisms:
  * backslash double-quote escape, which is flawed because double backslash is not also an escape;
  * HTML escapes surrounded by angle brackets.
  we use the double-quote syntax when no escaping is required because it looks nicer,
  but rely on HTML escaping for everything else.
  See https://www.graphviz.org/doc/info/lang.html.
  '''
  if isinstance(name, (int, float)): return str(name)
  assert isinstance(name, str), name
  if dot_bare_id_re.fullmatch(name) and name not in dot_keywords: return name
  if dot_quotable_id_re.fullmatch(name): return f'"{name}"'
  return f'<{html_escape(name)}>'


def write_dot_digraph_adjacency_contents(f: TextIO, adjacency:GraphvizAdjacency) -> None:
  if isinstance(adjacency, dict): adjacency = adjacency.items()
  for src, dsts in adjacency: # type: ignore
    if isinstance(dsts, dict): dsts = dsts.items()
    src_quoted = dot_id_quote(src)
    for dst in dsts:
      attrs_str = ''
      if isinstance(dst, tuple):
        dst, attrs = dst
        if not isinstance(attrs, dict): raise ValueError(attrs)
        attrs_str = fmt_dot_attrs(attrs)
      f.write(f'{src_quoted} -> {dot_id_quote(dst)}{attrs_str};\n')


def write_dot_digraph_adjacency(f: TextIO, adjacency:GraphvizAdjacency, nodes:GraphvizNodes=None, **kwargs) -> None:
  label = dot_id_quote(kwargs.get('label', ''))
  if label: label += ' '
  f.write(f'strict digraph {label}{{\n')
  # Graph attributes.
  for k, v in kwargs.items():
    validator = graph_prop_validators[k]
    if not validator(v): raise ValueError(f'value for {k} failed validation: {v!r}')
    f.write(f'{k}={dot_id_quote(v)};\n')
  # Nodes.
  if nodes:
    for node, attrs in nodes.items():
      f.write(f'{dot_id_quote(node)}{fmt_dot_attrs(attrs)};\n')
  # Edges.
  write_dot_digraph_adjacency_contents(f, adjacency)
  f.write('}\n')


def fmt_dot_attrs(attrs:Optional[GraphvizAttrs]) -> str:
  if not attrs: return ''
  s = ', '.join(f'{k}={dot_id_quote(v)}' for k, v in attrs.items())
  return f' [{s}]'

def out_dot_digraph_adjacency(adjacency:GraphvizAdjacency, **kwargs) -> None:
  write_dot_digraph_adjacency(stdout, adjacency=adjacency, **kwargs)


graph_prop_validators: Dict[str, Callable[[Any], bool]] = {
  'label': lambda v: isinstance(v, str),
  'rankdir': lambda v: (v in {'TB', 'BT', 'LR', 'RL'}),
}



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'print file hashes. first argument is the format; remaining args are the paths.'

import hashlib
from argparse import ArgumentParser
from base64 import b16encode, b32encode, b64encode, urlsafe_b64encode
from typing import Any, ByteString, Callable, Dict, List, Optional, Tuple, TypeVar

from ..encodings import enc_lep62
from ..io import errSL


_ByteString = TypeVar('_ByteString', ByteString, bytes, bytearray, memoryview) # Hack around the typeshed defs from base64.
_Encoder = Callable[[_ByteString], bytes]

hashes:Dict[str,Any] = {
  'blake2b'   : hashlib.blake2b,
  'blake2s'   : hashlib.blake2s,
  'md5'       : hashlib.md5,
  'sha1'      : hashlib.sha1,
  'sha224'    : hashlib.sha224,
  'sha256'    : hashlib.sha256,
  'sha3_224'  : hashlib.sha3_224,
  'sha3_256'  : hashlib.sha3_256,
  'sha3_384'  : hashlib.sha3_384,
  'sha3_512'  : hashlib.sha3_512,
  'sha384'    : hashlib.sha384,
  'sha512'    : hashlib.sha512,
  'shake_128' : hashlib.shake_128,
  'shake_256' : hashlib.shake_256,
}

variable_size_hashes = { 'blake2b', 'blake2s', 'shake_128', 'shake_256' }
variable_constructor_hashes = { 'blake2b', 'blake2s' }
variable_digest_arg_hashes = { 'shake_128', 'shake_256' }

hash_docs_str = ', '.join(hashes)


def main() -> None:
  parser = ArgumentParser(description='Count lines of source code.')
  parser.add_argument('-hash', default='blake2b', help=f'Hash algorithm to use: {hash_docs_str}.')
  parser.add_argument('-size', default=32, type=int, help='Digest size.')
  parser.add_argument('-lep62', action='store_true', help='Show lep62 result (default).')
  parser.add_argument('-b16', action='store_true', help='Show base16 result.')
  parser.add_argument('-b32', action='store_true', help='Show base32 result.')
  parser.add_argument('-b64', action='store_true', help='Show base64 result.')
  parser.add_argument('-url64', action='store_true', help='Show URL-safe base64 result.')

  parser.add_argument('paths', nargs='+', help='files to hash.')
  args = parser.parse_args()
  encoders:List[Tuple[str,_Encoder]] = []
  if args.lep62:  encoders.append(('lep62', enc_lep62))
  if args.b16:    encoders.append(('b16', b16encode))
  if args.b32:    encoders.append(('b32', b32encode))
  if args.b64:    encoders.append(('b64', b64encode))
  if args.url64:  encoders.append(('url64', urlsafe_b64encode))
  if not encoders:
    encoders.append(('lep62', enc_lep62))

  try:
    hash_class = hashes[args.hash]
  except KeyError:
    errSL('error: invalid hash name:', args.hash)
    errSL('note: available hash functions:', *hashes)
    exit(1)

  hash_size = args.size
  if hash_size is None and args.hash in variable_digest_arg_hashes:
    exit(f'error: `-size` must be specified for hash function "{args.hash}".')

  size_arg = {'digest_size':args.size} if args.hash in variable_constructor_hashes else {}

  path_width = min(64, max(len(p) for p in args.paths))

  for path in args.paths:
    hasher = hash_class(**size_arg)
    hs = hash_size if args.hash in variable_digest_arg_hashes else None
    d = digest(hasher, path, hash_size=hs)
    msgs = []
    for label, encoder in encoders:
      msgs.append(f'{label}:{encoder(d).decode()}')
    print(f'{args.hash} {path:{path_width}}', *msgs)


def digest(hasher:Any, path:str, hash_size:Optional[int]) -> bytes:
  hash_chunk_size = 1 << 16
  #^ a quick timing experiment suggested that chunk sizes larger than this are not faster.
  try: f = open(path, 'rb')
  except IsADirectoryError: exit(f'expected a file but found a directory: {path}')
  while True:
    chunk = f.read(hash_chunk_size)
    if not chunk: break
    hasher.update(chunk)
  if hash_size: return hasher.digest(hash_size) # type: ignore
  else: return hasher.digest() # type: ignore


if __name__ == '__main__': main()
